#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any, TypedDict
from unittest import mock

import pytest
from botocore.waiter import Waiter

from airflow.exceptions import AirflowProviderDeprecationWarning, TaskDeferred
from airflow.providers.amazon.aws.hooks.eks import ClusterStates, EksHook
from airflow.providers.amazon.aws.operators.eks import (
    EksCreateClusterOperator,
    EksCreateFargateProfileOperator,
    EksCreateNodegroupOperator,
    EksDeleteClusterOperator,
    EksDeleteFargateProfileOperator,
    EksDeleteNodegroupOperator,
    EksPodOperator,
)
from airflow.providers.amazon.aws.triggers.eks import (
    EksCreateFargateProfileTrigger,
    EksCreateNodegroupTrigger,
    EksDeleteFargateProfileTrigger,
)
from airflow.providers.cncf.kubernetes.utils.pod_manager import OnFinishAction

from unit.amazon.aws.utils.eks_test_constants import (
    NODEROLE_ARN,
    POD_EXECUTION_ROLE_ARN,
    RESOURCES_VPC_CONFIG,
    ROLE_ARN,
    SELECTORS,
    SUBNET_IDS,
    TASK_ID,
)
from unit.amazon.aws.utils.eks_test_utils import convert_keys
from unit.amazon.aws.utils.test_template_fields import validate_template_fields
from unit.amazon.aws.utils.test_waiter import assert_expected_waiter_type

CLUSTER_NAME = "cluster1"
NODEGROUP_NAME = "nodegroup1"
FARGATE_PROFILE_NAME = "fargate_profile1"
DESCRIBE_CLUSTER_RESULT = f'{{"cluster": "{CLUSTER_NAME}"}}'
DESCRIBE_NODEGROUP_RESULT = f'{{"nodegroup": "{NODEGROUP_NAME}"}}'
EMPTY_CLUSTER = '{"cluster": {}}'
EMPTY_NODEGROUP = '{"nodegroup": {}}'
NAME_LIST = ["foo", "bar", "baz", "qux"]
CREATE_CLUSTER_KWARGS = {"version": "1.22"}
CREATE_FARGATE_PROFILE_KWARGS = {"tags": {"hello": "world"}}
CREATE_NODEGROUP_KWARGS = {
    "capacityType": "ON_DEMAND",
    "instanceTypes": "t3.large",
}


class ClusterParams(TypedDict):
    cluster_name: str
    cluster_role_arn: str
    resources_vpc_config: dict[Any, Any]


class NodeGroupParams(TypedDict):
    nodegroup_name: str
    nodegroup_role_arn: str


class BaseFargateProfileParams(TypedDict):
    fargate_profile_name: str
    fargate_pod_execution_role_arn: str
    fargate_selectors: list[Any]


class CreateFargateProfileParams(TypedDict):
    cluster_name: str
    pod_execution_role_arn: str
    selectors: list[Any]
    fargate_profile_name: str


class CreateNodegroupParams(TypedDict):
    cluster_name: str
    nodegroup_name: str
    nodegroup_subnets: list[str]
    nodegroup_role_arn: str


class TestEksCreateClusterOperator:
    def setup_method(self) -> None:
        # Parameters which are needed to create a cluster.
        self.create_cluster_params = ClusterParams(
            cluster_name=CLUSTER_NAME,
            cluster_role_arn=ROLE_ARN[1],
            resources_vpc_config=RESOURCES_VPC_CONFIG[1],
        )
        self.nodegroup_setup()
        self.fargate_profile_setup()

    def nodegroup_setup(self) -> None:
        # Parameters which are added to the cluster parameters
        # when creating both the cluster and nodegroup together.
        self.base_nodegroup_params = NodeGroupParams(
            nodegroup_name=NODEGROUP_NAME,
            nodegroup_role_arn=NODEROLE_ARN[1],
        )

        # Parameters expected to be passed in the CreateNodegroup hook call.
        self.create_nodegroup_params = dict(
            **self.base_nodegroup_params,
            cluster_name=CLUSTER_NAME,
            subnets=SUBNET_IDS,
        )

        self.create_cluster_operator_with_nodegroup = EksCreateClusterOperator(
            task_id=TASK_ID,
            **self.create_cluster_params,
            **self.base_nodegroup_params,
        )

    def fargate_profile_setup(self) -> None:
        # Parameters which are added to the cluster parameters
        # when creating both the cluster and Fargate profile together.
        self.base_fargate_profile_params = BaseFargateProfileParams(
            fargate_profile_name=FARGATE_PROFILE_NAME,
            fargate_pod_execution_role_arn=POD_EXECUTION_ROLE_ARN[1],
            fargate_selectors=SELECTORS[1],
        )

        # Parameters expected to be passed in the CreateFargateProfile hook call.
        self.create_fargate_profile_params = dict(
            **self.base_fargate_profile_params,
            cluster_name=CLUSTER_NAME,
        )

        self.create_cluster_operator_with_fargate_profile = EksCreateClusterOperator(
            task_id=TASK_ID,
            **self.create_cluster_params,
            **self.base_fargate_profile_params,
            compute="fargate",
        )

    @pytest.mark.parametrize(
        "create_cluster_kwargs",
        [
            pytest.param(None, id="without cluster kwargs"),
            pytest.param(CREATE_CLUSTER_KWARGS, id="with cluster kwargs"),
        ],
    )
    @mock.patch.object(EksHook, "create_cluster")
    @mock.patch.object(EksHook, "create_nodegroup")
    def test_execute_create_cluster(self, mock_create_nodegroup, mock_create_cluster, create_cluster_kwargs):
        op_kwargs = {**self.create_cluster_params, "compute": None}
        if create_cluster_kwargs:
            op_kwargs["create_cluster_kwargs"] = create_cluster_kwargs
            parameters = {**self.create_cluster_params, **create_cluster_kwargs}
        else:
            assert "create_cluster_kwargs" not in op_kwargs
            parameters = self.create_cluster_params

        operator = EksCreateClusterOperator(task_id=TASK_ID, **op_kwargs)
        operator.execute({})
        mock_create_cluster.assert_called_with(**convert_keys(parameters))
        mock_create_nodegroup.assert_not_called()

    @pytest.mark.parametrize(
        "create_cluster_kwargs",
        [
            pytest.param(None, id="without cluster kwargs"),
            pytest.param(CREATE_CLUSTER_KWARGS, id="with cluster kwargs"),
        ],
    )
    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "create_cluster")
    @mock.patch.object(EksHook, "create_nodegroup")
    def test_execute_create_cluster_with_wait(
        self, mock_create_nodegroup, mock_create_cluster, mock_waiter, create_cluster_kwargs
    ):
        op_kwargs = {**self.create_cluster_params, "compute": None}
        if create_cluster_kwargs:
            op_kwargs["create_cluster_kwargs"] = create_cluster_kwargs
            parameters = {**self.create_cluster_params, **create_cluster_kwargs}
        else:
            assert "create_cluster_kwargs" not in op_kwargs
            parameters = self.create_cluster_params

        operator = EksCreateClusterOperator(task_id=TASK_ID, **op_kwargs, wait_for_completion=True)
        operator.execute({})
        mock_create_cluster.assert_called_with(**convert_keys(parameters))
        mock_create_nodegroup.assert_not_called()
        mock_waiter.assert_called_once_with(
            mock.ANY,
            name=CLUSTER_NAME,
            WaiterConfig={"Delay": mock.ANY, "MaxAttempts": mock.ANY},
        )
        assert_expected_waiter_type(mock_waiter, "ClusterActive")

    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "get_cluster_state")
    @mock.patch.object(EksHook, "create_cluster")
    @mock.patch.object(EksHook, "create_nodegroup")
    def test_execute_when_called_with_nodegroup_creates_both(
        self, mock_create_nodegroup, mock_create_cluster, mock_cluster_state, mock_waiter
    ):
        mock_cluster_state.return_value = ClusterStates.ACTIVE

        self.create_cluster_operator_with_nodegroup.execute({})

        mock_create_cluster.assert_called_once_with(**convert_keys(self.create_cluster_params))
        mock_create_nodegroup.assert_called_once_with(**convert_keys(self.create_nodegroup_params))
        mock_waiter.assert_called_once_with(
            mock.ANY,
            name=CLUSTER_NAME,
            WaiterConfig={"Delay": mock.ANY, "MaxAttempts": mock.ANY},
        )
        assert_expected_waiter_type(mock_waiter, "ClusterActive")

    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "get_cluster_state")
    @mock.patch.object(EksHook, "create_cluster")
    @mock.patch.object(EksHook, "create_nodegroup")
    def test_execute_with_wait_when_called_with_nodegroup_creates_both(
        self, mock_create_nodegroup, mock_create_cluster, mock_cluster_state, mock_waiter
    ):
        mock_cluster_state.return_value = ClusterStates.ACTIVE
        self.create_cluster_operator_with_nodegroup.wait_for_completion = True

        self.create_cluster_operator_with_nodegroup.execute({})

        mock_create_cluster.assert_called_once_with(**convert_keys(self.create_cluster_params))
        mock_create_nodegroup.assert_called_once_with(**convert_keys(self.create_nodegroup_params))
        # Calls waiter once for the cluster and once for the nodegroup.
        assert mock_waiter.call_count == 2
        mock_waiter.assert_called_with(
            mock.ANY,
            clusterName=CLUSTER_NAME,
            nodegroupName=NODEGROUP_NAME,
            WaiterConfig={"MaxAttempts": mock.ANY},
        )
        assert_expected_waiter_type(mock_waiter, "NodegroupActive")

    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "get_cluster_state")
    @mock.patch.object(EksHook, "create_cluster")
    @mock.patch.object(EksHook, "create_fargate_profile")
    def test_execute_when_called_with_fargate_creates_both(
        self, mock_create_fargate_profile, mock_create_cluster, mock_cluster_state, mock_waiter
    ):
        mock_cluster_state.return_value = ClusterStates.ACTIVE

        self.create_cluster_operator_with_fargate_profile.execute({})

        mock_create_cluster.assert_called_once_with(**convert_keys(self.create_cluster_params))
        mock_create_fargate_profile.assert_called_once_with(
            **convert_keys(self.create_fargate_profile_params)
        )
        mock_waiter.assert_called_once_with(
            mock.ANY,
            name=CLUSTER_NAME,
            WaiterConfig={"Delay": mock.ANY, "MaxAttempts": mock.ANY},
        )
        assert_expected_waiter_type(mock_waiter, "ClusterActive")

    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "get_cluster_state")
    @mock.patch.object(EksHook, "create_cluster")
    @mock.patch.object(EksHook, "create_fargate_profile")
    def test_execute_with_wait_when_called_with_fargate_creates_both(
        self, mock_create_fargate_profile, mock_create_cluster, mock_cluster_state, mock_waiter
    ):
        mock_cluster_state.return_value = ClusterStates.ACTIVE
        self.create_cluster_operator_with_fargate_profile.wait_for_completion = True

        self.create_cluster_operator_with_fargate_profile.execute({})

        mock_create_cluster.assert_called_once_with(**convert_keys(self.create_cluster_params))
        mock_create_fargate_profile.assert_called_once_with(
            **convert_keys(self.create_fargate_profile_params)
        )
        # Calls waiter once for the cluster and once for the nodegroup.
        assert mock_waiter.call_count == 2
        mock_waiter.assert_called_with(
            mock.ANY,
            clusterName=CLUSTER_NAME,
            fargateProfileName=FARGATE_PROFILE_NAME,
            WaiterConfig={"MaxAttempts": mock.ANY},
        )
        assert_expected_waiter_type(mock_waiter, "FargateProfileActive")

    def test_invalid_compute_value(self):
        invalid_compute = EksCreateClusterOperator(
            task_id=TASK_ID,
            **self.create_cluster_params,
            compute="infinite",
        )

        with pytest.raises(ValueError, match="Provided compute type is not supported."):
            invalid_compute.execute({})

    def test_nodegroup_compute_missing_nodegroup_role_arn(self):
        missing_nodegroup_role_arn = EksCreateClusterOperator(
            task_id=TASK_ID,
            **self.create_cluster_params,
            compute="nodegroup",
        )

        with pytest.raises(
            ValueError,
            match="Creating an Amazon EKS managed node groups requires nodegroup_role_arn to be passed in.",
        ):
            missing_nodegroup_role_arn.execute({})

    def test_fargate_compute_missing_fargate_pod_execution_role_arn(self):
        missing_fargate_pod_execution_role_arn = EksCreateClusterOperator(
            task_id=TASK_ID,
            **self.create_cluster_params,
            compute="fargate",
        )

        with pytest.raises(
            ValueError,
            match="Creating an AWS Fargate profiles requires fargate_pod_execution_role_arn to be passed in.",
        ):
            missing_fargate_pod_execution_role_arn.execute({})

    def test_init_with_region(self):
        with pytest.warns(AirflowProviderDeprecationWarning) as m:
            m.operator = EksCreateClusterOperator(
                task_id=TASK_ID,
                **self.create_cluster_params,
                compute=None,
                region="us-east-2",
            )
        assert m.operator.region_name == "us-east-2"

    @mock.patch.object(EksHook, "create_cluster")
    def test_eks_create_cluster_short_circuit_early(self, mock_create_cluster, caplog):
        mock_create_cluster.return_value = None
        eks_create_cluster_operator = EksCreateClusterOperator(
            task_id=TASK_ID,
            **self.create_cluster_params,
            compute=None,
            wait_for_completion=False,
            deferrable=False,
        )
        eks_create_cluster_operator.execute({})
        assert len(caplog.records) == 0

    @mock.patch.object(EksHook, "create_cluster")
    def test_eks_create_cluster_with_deferrable(self, mock_create_cluster, caplog):
        mock_create_cluster.return_value = None

        eks_create_cluster_operator = EksCreateClusterOperator(
            task_id=TASK_ID,
            **self.create_cluster_params,
            compute=None,
            wait_for_completion=False,
            deferrable=True,
        )
        with pytest.raises(TaskDeferred):
            eks_create_cluster_operator.execute({})
        assert "Waiting for EKS Cluster to provision. This will take some time." in caplog.messages

    def test_template_fields(self):
        op = EksCreateClusterOperator(
            task_id=TASK_ID, **self.create_cluster_params, compute="fargate", region_name="us-east-1"
        )

        validate_template_fields(op)


class TestEksCreateFargateProfileOperator:
    def setup_method(self) -> None:
        self.create_fargate_profile_params = CreateFargateProfileParams(  # type: ignore
            cluster_name=CLUSTER_NAME,
            pod_execution_role_arn=POD_EXECUTION_ROLE_ARN[1],
            selectors=SELECTORS[1],
            fargate_profile_name=FARGATE_PROFILE_NAME,
        )

    @pytest.mark.parametrize(
        "create_fargate_profile_kwargs",
        [
            pytest.param(None, id="without fargate profile kwargs"),
            pytest.param(CREATE_FARGATE_PROFILE_KWARGS, id="with fargate profile kwargs"),
        ],
    )
    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "create_fargate_profile")
    def test_execute_when_fargate_profile_does_not_already_exist(
        self, mock_create_fargate_profile, mock_waiter, create_fargate_profile_kwargs
    ):
        op_kwargs = {**self.create_fargate_profile_params}
        if create_fargate_profile_kwargs:
            op_kwargs["create_fargate_profile_kwargs"] = create_fargate_profile_kwargs
            parameters = {**self.create_fargate_profile_params, **create_fargate_profile_kwargs}
        else:
            assert "create_fargate_profile_kwargs" not in op_kwargs
            parameters = self.create_fargate_profile_params

        operator = EksCreateFargateProfileOperator(task_id=TASK_ID, **op_kwargs)
        operator.execute({})
        mock_create_fargate_profile.assert_called_with(**convert_keys(parameters))
        mock_waiter.assert_not_called()

    @pytest.mark.parametrize(
        "create_fargate_profile_kwargs",
        [
            pytest.param(None, id="without fargate profile kwargs"),
            pytest.param(CREATE_FARGATE_PROFILE_KWARGS, id="with fargate profile kwargs"),
        ],
    )
    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "create_fargate_profile")
    def test_execute_with_wait_when_fargate_profile_does_not_already_exist(
        self, mock_create_fargate_profile, mock_waiter, create_fargate_profile_kwargs
    ):
        op_kwargs = {**self.create_fargate_profile_params}
        if create_fargate_profile_kwargs:
            op_kwargs["create_fargate_profile_kwargs"] = create_fargate_profile_kwargs
            parameters = {**self.create_fargate_profile_params, **create_fargate_profile_kwargs}
        else:
            assert "create_fargate_profile_kwargs" not in op_kwargs
            parameters = self.create_fargate_profile_params

        operator = EksCreateFargateProfileOperator(task_id=TASK_ID, **op_kwargs, wait_for_completion=True)
        operator.execute({})
        mock_create_fargate_profile.assert_called_with(**convert_keys(parameters))
        mock_waiter.assert_called_with(
            mock.ANY,
            clusterName=CLUSTER_NAME,
            fargateProfileName=FARGATE_PROFILE_NAME,
            WaiterConfig={"MaxAttempts": mock.ANY},
        )
        assert_expected_waiter_type(mock_waiter, "FargateProfileActive")

    @mock.patch.object(EksHook, "create_fargate_profile")
    def test_create_fargate_profile_deferrable(self, _):
        op_kwargs = {**self.create_fargate_profile_params}
        operator = EksCreateFargateProfileOperator(
            task_id=TASK_ID,
            **op_kwargs,
            deferrable=True,
        )
        with pytest.raises(TaskDeferred) as exc:
            operator.execute({})
        assert isinstance(exc.value.trigger, EksCreateFargateProfileTrigger), (
            "Trigger is not a EksCreateFargateProfileTrigger"
        )

    def test_template_fields(self):
        op = EksCreateFargateProfileOperator(task_id=TASK_ID, **self.create_fargate_profile_params)

        validate_template_fields(op)

    def test_init_with_region(self):
        with pytest.warns(AirflowProviderDeprecationWarning) as m:
            m.operator = EksCreateFargateProfileOperator(
                task_id=TASK_ID,
                **self.create_fargate_profile_params,
                region="us-east-2",
            )
        assert m.operator.region_name == "us-east-2"


class TestEksCreateNodegroupOperator:
    def setup_method(self) -> None:
        self.create_nodegroup_params = CreateNodegroupParams(
            cluster_name=CLUSTER_NAME,
            nodegroup_name=NODEGROUP_NAME,
            nodegroup_subnets=SUBNET_IDS,
            nodegroup_role_arn=NODEROLE_ARN[1],
        )

    @pytest.mark.parametrize(
        "create_nodegroup_kwargs",
        [
            pytest.param(None, id="without nodegroup kwargs"),
            pytest.param(CREATE_NODEGROUP_KWARGS, id="with nodegroup kwargs"),
        ],
    )
    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "create_nodegroup")
    def test_execute_when_nodegroup_does_not_already_exist(
        self, mock_create_nodegroup, mock_waiter, create_nodegroup_kwargs
    ):
        op_kwargs = {**self.create_nodegroup_params}
        if create_nodegroup_kwargs:
            op_kwargs["create_nodegroup_kwargs"] = create_nodegroup_kwargs
            parameters = {**self.create_nodegroup_params, **create_nodegroup_kwargs}
        else:
            assert "create_nodegroup_params" not in op_kwargs
            parameters = self.create_nodegroup_params

        operator = EksCreateNodegroupOperator(task_id=TASK_ID, **op_kwargs)
        operator.execute({})
        mock_create_nodegroup.assert_called_with(**convert_keys(parameters))
        mock_waiter.assert_not_called()

        @pytest.mark.parametrize(
            "create_nodegroup_kwargs",
            [
                pytest.param(None, id="without nodegroup kwargs"),
                pytest.param(CREATE_NODEGROUP_KWARGS, id="with nodegroup kwargs"),
            ],
        )
        @mock.patch.object(Waiter, "wait")
        @mock.patch.object(EksHook, "create_nodegroup")
        def test_execute_with_wait_when_nodegroup_does_not_already_exist(
            self, mock_create_nodegroup, mock_waiter, create_nodegroup_kwargs
        ):
            op_kwargs = {**self.create_nodegroup_params}
            if create_nodegroup_kwargs:
                op_kwargs["create_nodegroup_kwargs"] = create_nodegroup_kwargs
                parameters = {**self.create_nodegroup_params, **create_nodegroup_kwargs}
            else:
                assert "create_nodegroup_params" not in op_kwargs
                parameters = self.create_nodegroup_params

            operator = EksCreateNodegroupOperator(task_id=TASK_ID, **op_kwargs, wait_for_completion=True)
            operator.execute({})
            mock_create_nodegroup.assert_called_with(**convert_keys(parameters))
            mock_waiter.assert_called_with(mock.ANY, clusterName=CLUSTER_NAME, nodegroupName=NODEGROUP_NAME)
            assert_expected_waiter_type(mock_waiter, "NodegroupActive")

    @mock.patch.object(EksHook, "create_nodegroup")
    def test_create_nodegroup_deferrable(self, mock_create_nodegroup):
        mock_create_nodegroup.return_value = True
        op_kwargs = {**self.create_nodegroup_params}
        operator = EksCreateNodegroupOperator(
            task_id=TASK_ID,
            **op_kwargs,
            deferrable=True,
        )
        with pytest.raises(TaskDeferred) as exc:
            operator.execute({})
        assert isinstance(exc.value.trigger, EksCreateNodegroupTrigger)

    def test_create_nodegroup_deferrable_versus_wait_for_completion(self):
        op_kwargs = {**self.create_nodegroup_params}
        operator = EksCreateNodegroupOperator(
            task_id=TASK_ID,
            **op_kwargs,
            deferrable=True,
            wait_for_completion=True,
        )
        assert operator.wait_for_completion is False
        operator = EksCreateNodegroupOperator(
            task_id=TASK_ID,
            **op_kwargs,
            deferrable=False,
            wait_for_completion=True,
        )
        assert operator.wait_for_completion is True

    def test_template_fields(self):
        op_kwargs = {**self.create_nodegroup_params}
        op = EksCreateNodegroupOperator(task_id=TASK_ID, **op_kwargs)

        validate_template_fields(op)

    def test_init_with_region(self):
        op_kwargs = {**self.create_nodegroup_params}

        with pytest.warns(AirflowProviderDeprecationWarning) as m:
            m.operator = EksCreateNodegroupOperator(
                task_id=TASK_ID,
                **op_kwargs,
                region="us-east-2",
            )
        assert m.operator.region_name == "us-east-2"


class TestEksDeleteClusterOperator:
    def setup_method(self) -> None:
        self.cluster_name: str = CLUSTER_NAME

        self.delete_cluster_operator = EksDeleteClusterOperator(
            task_id=TASK_ID, cluster_name=self.cluster_name
        )

    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "list_nodegroups")
    @mock.patch.object(EksHook, "delete_cluster")
    def test_existing_cluster_not_in_use(self, mock_delete_cluster, mock_list_nodegroups, mock_waiter):
        mock_list_nodegroups.return_value = []
        self.delete_cluster_operator.execute({})
        mock_delete_cluster.assert_called_once_with(name=self.cluster_name)
        mock_waiter.assert_not_called()

    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "list_nodegroups")
    @mock.patch.object(EksHook, "delete_cluster")
    def test_existing_cluster_not_in_use_with_wait(
        self, mock_delete_cluster, mock_list_nodegroups, mock_waiter
    ):
        mock_list_nodegroups.return_value = []
        self.delete_cluster_operator.wait_for_completion = True

        self.delete_cluster_operator.execute({})

        mock_list_nodegroups.assert_not_called()
        mock_delete_cluster.assert_called_once_with(name=self.cluster_name)
        mock_waiter.assert_called_with(mock.ANY, name=CLUSTER_NAME)
        assert_expected_waiter_type(mock_waiter, "ClusterDeleted")

    def test_eks_delete_cluster_operator_with_deferrable(self):
        self.delete_cluster_operator.deferrable = True
        with pytest.raises(TaskDeferred):
            self.delete_cluster_operator.execute({})

    def test_template_fields(self):
        validate_template_fields(self.delete_cluster_operator)

    def test_init_with_region(self):
        with pytest.warns(AirflowProviderDeprecationWarning) as m:
            m.operator = EksDeleteClusterOperator(
                task_id=TASK_ID, cluster_name=self.cluster_name, region="us-east-2"
            )
        assert m.operator.region_name == "us-east-2"


class TestEksDeleteNodegroupOperator:
    def setup_method(self) -> None:
        self.cluster_name: str = CLUSTER_NAME
        self.nodegroup_name: str = NODEGROUP_NAME

        self.delete_nodegroup_operator = EksDeleteNodegroupOperator(
            task_id=TASK_ID, cluster_name=self.cluster_name, nodegroup_name=self.nodegroup_name
        )

    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "delete_nodegroup")
    def test_existing_nodegroup(self, mock_delete_nodegroup, mock_waiter):
        self.delete_nodegroup_operator.execute({})

        mock_delete_nodegroup.assert_called_once_with(
            clusterName=self.cluster_name, nodegroupName=self.nodegroup_name
        )
        mock_waiter.assert_not_called()

    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "delete_nodegroup")
    def test_existing_nodegroup_with_wait(self, mock_delete_nodegroup, mock_waiter):
        self.delete_nodegroup_operator.wait_for_completion = True

        self.delete_nodegroup_operator.execute({})

        mock_delete_nodegroup.assert_called_once_with(
            clusterName=self.cluster_name, nodegroupName=self.nodegroup_name
        )
        mock_waiter.assert_called_with(mock.ANY, clusterName=CLUSTER_NAME, nodegroupName=NODEGROUP_NAME)
        assert_expected_waiter_type(mock_waiter, "NodegroupDeleted")

    def test_template_fields(self):
        validate_template_fields(self.delete_nodegroup_operator)

    def test_init_with_region(self):
        with pytest.warns(AirflowProviderDeprecationWarning) as m:
            m.operator = EksDeleteNodegroupOperator(
                task_id=TASK_ID,
                cluster_name=self.cluster_name,
                nodegroup_name=self.nodegroup_name,
                region="us-east-2",
            )
        assert m.operator.region_name == "us-east-2"


class TestEksDeleteFargateProfileOperator:
    def setup_method(self) -> None:
        self.cluster_name: str = CLUSTER_NAME
        self.fargate_profile_name: str = FARGATE_PROFILE_NAME

        self.delete_fargate_profile_operator = EksDeleteFargateProfileOperator(
            task_id=TASK_ID, cluster_name=self.cluster_name, fargate_profile_name=self.fargate_profile_name
        )

    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "delete_fargate_profile")
    def test_existing_fargate_profile(self, mock_delete_fargate_profile, mock_waiter):
        self.delete_fargate_profile_operator.execute({})

        mock_delete_fargate_profile.assert_called_once_with(
            clusterName=self.cluster_name, fargateProfileName=self.fargate_profile_name
        )
        mock_waiter.assert_not_called()

    @mock.patch.object(Waiter, "wait")
    @mock.patch.object(EksHook, "delete_fargate_profile")
    def test_existing_fargate_profile_with_wait(self, mock_delete_fargate_profile, mock_waiter):
        self.delete_fargate_profile_operator.wait_for_completion = True

        self.delete_fargate_profile_operator.execute({})

        mock_delete_fargate_profile.assert_called_once_with(
            clusterName=self.cluster_name, fargateProfileName=self.fargate_profile_name
        )
        mock_waiter.assert_called_with(
            mock.ANY,
            clusterName=CLUSTER_NAME,
            fargateProfileName=FARGATE_PROFILE_NAME,
            WaiterConfig={"Delay": 30, "MaxAttempts": 60},
        )
        assert_expected_waiter_type(mock_waiter, "FargateProfileDeleted")

    @mock.patch.object(EksHook, "delete_fargate_profile")
    def test_delete_fargate_profile_deferrable(self, _):
        self.delete_fargate_profile_operator.deferrable = True

        with pytest.raises(TaskDeferred) as exc:
            self.delete_fargate_profile_operator.execute({})
        assert isinstance(exc.value.trigger, EksDeleteFargateProfileTrigger), (
            "Trigger is not a EksDeleteFargateProfileTrigger"
        )

    def test_template_fields(self):
        validate_template_fields(self.delete_fargate_profile_operator)

    def test_init_with_region(self):
        with pytest.warns(AirflowProviderDeprecationWarning) as m:
            m.operator = EksDeleteFargateProfileOperator(
                task_id=TASK_ID,
                cluster_name=self.cluster_name,
                fargate_profile_name=self.fargate_profile_name,
                region="us-east-2",
            )
        assert m.operator.region_name == "us-east-2"


class TestEksPodOperator:
    @mock.patch("airflow.providers.cncf.kubernetes.operators.pod.KubernetesPodOperator.execute")
    @mock.patch("airflow.providers.amazon.aws.hooks.eks.EksHook.generate_config_file")
    @mock.patch("airflow.providers.amazon.aws.hooks.eks.EksHook.__init__", return_value=None)
    def test_existing_nodegroup(
        self, mock_eks_hook, mock_generate_config_file, mock_k8s_pod_operator_execute
    ):
        ti_context = mock.MagicMock(name="ti_context")

        op = EksPodOperator(
            task_id="run_pod",
            pod_name="run_pod",
            cluster_name=CLUSTER_NAME,
            image="amazon/aws-cli:latest",
            cmds=["sh", "-c", "ls"],
            labels={"demo": "hello_world"},
            get_logs=True,
            # Delete the pod when it reaches its final state, or the execution is interrupted.
            on_finish_action="delete_pod",
        )
        op_return_value = op.execute(ti_context)
        mock_k8s_pod_operator_execute.assert_called_once_with(ti_context)
        mock_eks_hook.assert_called_once_with(aws_conn_id="aws_default", region_name=None)
        mock_generate_config_file.assert_called_once_with(
            eks_cluster_name=CLUSTER_NAME, pod_namespace="default"
        )
        assert mock_k8s_pod_operator_execute.return_value == op_return_value
        assert mock_generate_config_file.return_value.__enter__.return_value == op.config_file

    @pytest.mark.parametrize(
        "compatible_kpo, kwargs, expected_attributes",
        [
            (
                True,
                {"on_finish_action": "delete_succeeded_pod"},
                {"on_finish_action": OnFinishAction.DELETE_SUCCEEDED_POD},
            ),
            (
                # test default
                True,
                {},
                {"on_finish_action": OnFinishAction.DELETE_POD},
            ),
            (
                # test default
                False,
                {},
                {},
            ),
        ],
    )
    def test_on_finish_action_handler(self, compatible_kpo, kwargs, expected_attributes):
        kpo_init_args_mock = mock.MagicMock(**{"parameters": ["on_finish_action"] if compatible_kpo else []})

        with mock.patch("inspect.signature", return_value=kpo_init_args_mock):
            op = EksPodOperator(
                task_id="run_pod",
                pod_name="run_pod",
                cluster_name=CLUSTER_NAME,
                image="amazon/aws-cli:latest",
                cmds=["sh", "-c", "ls"],
                labels={"demo": "hello_world"},
                get_logs=True,
                **kwargs,
            )
            for expected_attr in expected_attributes:
                assert op.__getattribute__(expected_attr) == expected_attributes[expected_attr]

    def test_template_fields(self):
        op = EksPodOperator(
            task_id="run_pod",
            pod_name="run_pod",
            cluster_name=CLUSTER_NAME,
            image="amazon/aws-cli:latest",
            cmds=["sh", "-c", "ls"],
            labels={"demo": "hello_world"},
            get_logs=True,
            on_finish_action="delete_pod",
        )

        validate_template_fields(op)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import importlib
from unittest.mock import ANY, Mock, patch

import pytest

from airflow.providers.amazon.version_compat import AIRFLOW_V_3_0_PLUS

if not AIRFLOW_V_3_0_PLUS:
    pytest.skip("AWS auth manager is only compatible with Airflow >= 3.0.0", allow_module_level=True)

from airflow.cli import cli_parser
from airflow.providers.amazon.aws.auth_manager.cli.avp_commands import init_avp, update_schema

from tests_common.test_utils.config import conf_vars

mock_boto3 = Mock()


@pytest.mark.db_test
class TestAvpCommands:
    def setup_method(self):
        mock_boto3.reset_mock()

    @classmethod
    def setup_class(cls):
        with conf_vars(
            {
                (
                    "core",
                    "auth_manager",
                ): "airflow.providers.amazon.aws.auth_manager.aws_auth_manager.AwsAuthManager"
            }
        ):
            importlib.reload(cli_parser)
            cls.arg_parser = cli_parser.get_parser()

    @pytest.mark.parametrize(
        "dry_run",
        [False, True],
    )
    @patch("airflow.providers.amazon.aws.auth_manager.cli.avp_commands._get_client")
    def test_init_avp_with_no_existing_resources(self, mock_get_client, dry_run):
        mock_get_client.return_value = mock_boto3

        policy_store_description = "test-policy-store"
        policy_store_id = "test-policy-store-id"

        paginator = Mock()
        paginator.paginate.return_value = []

        mock_boto3.get_paginator.return_value = paginator
        mock_boto3.create_policy_store.return_value = {"policyStoreId": policy_store_id}

        with conf_vars({("database", "check_migrations"): "False"}):
            params = [
                "aws-auth-manager",
                "init-avp",
                "--policy-store-description",
                policy_store_description,
            ]
            if dry_run:
                params.append("--dry-run")
            init_avp(self.arg_parser.parse_args(params))

        if dry_run:
            mock_boto3.create_policy_store.assert_not_called()
            mock_boto3.put_schema.assert_not_called()
        else:
            mock_boto3.create_policy_store.assert_called_once_with(
                validationSettings={
                    "mode": "STRICT",
                },
                description=policy_store_description,
            )
            mock_boto3.put_schema.assert_called_once_with(
                policyStoreId=policy_store_id,
                definition={
                    "cedarJson": ANY,
                },
            )

    @pytest.mark.parametrize(
        "dry_run",
        [False, True],
    )
    @patch("airflow.providers.amazon.aws.auth_manager.cli.avp_commands._get_client")
    def test_init_avp_with_existing_resources(self, mock_get_client, dry_run):
        mock_get_client.return_value = mock_boto3

        policy_store_description = "test-policy-store"
        policy_store_id = "test-policy-store-id"

        paginator = Mock()
        paginator.paginate.return_value = [
            {"policyStores": [{"description": policy_store_description, "policyStoreId": policy_store_id}]}
        ]

        mock_boto3.get_paginator.return_value = paginator

        with conf_vars({("database", "check_migrations"): "False"}):
            params = [
                "aws-auth-manager",
                "init-avp",
                "--policy-store-description",
                policy_store_description,
            ]
            if dry_run:
                params.append("--dry-run")
            init_avp(self.arg_parser.parse_args(params))

        mock_boto3.create_policy_store.assert_not_called()
        mock_boto3.update_policy_store.assert_not_called()
        mock_boto3.put_schema.assert_not_called()

    @pytest.mark.parametrize(
        "dry_run",
        [False, True],
    )
    @patch("airflow.providers.amazon.aws.auth_manager.cli.avp_commands._get_client")
    def test_update_schema(self, mock_get_client, dry_run):
        mock_get_client.return_value = mock_boto3

        policy_store_id = "test-policy-store-id"

        with conf_vars({("database", "check_migrations"): "False"}):
            params = [
                "aws-auth-manager",
                "update-avp-schema",
                "--policy-store-id",
                policy_store_id,
            ]
            if dry_run:
                params.append("--dry-run")
            update_schema(self.arg_parser.parse_args(params))

        if dry_run:
            mock_boto3.put_schema.assert_not_called()
        else:
            mock_boto3.put_schema.assert_called_once_with(
                policyStoreId=policy_store_id,
                definition={
                    "cedarJson": ANY,
                },
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import MagicMock

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.links.emr import (
    EmrClusterLink,
    EmrLogsLink,
    EmrServerlessCloudWatchLogsLink,
    EmrServerlessDashboardLink,
    EmrServerlessLogsLink,
    EmrServerlessS3LogsLink,
    get_log_uri,
    get_serverless_dashboard_url,
)
from airflow.providers.amazon.version_compat import AIRFLOW_V_3_0_PLUS

from unit.amazon.aws.links.test_base_aws import BaseAwsLinksTestCase

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk.execution_time.comms import XComResult


class TestEmrClusterLink(BaseAwsLinksTestCase):
    link_class = EmrClusterLink

    def test_extra_link(self, mock_supervisor_comms):
        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.get_message.return_value = XComResult(
                key=self.link_class.key,
                value={
                    "region_name": "us-west-1",
                    "aws_domain": self.link_class.get_aws_domain("aws"),
                    "aws_partition": "aws",
                    "job_flow_id": "j-TEST-FLOW-ID",
                },
            )
        self.assert_extra_link_url(
            expected_url=(
                "https://console.aws.amazon.com/emr/home?region=us-west-1#/clusterDetails/j-TEST-FLOW-ID"
            ),
            region_name="us-west-1",
            aws_partition="aws",
            job_flow_id="j-TEST-FLOW-ID",
        )


@pytest.mark.parametrize(
    "cluster_info, expected_uri",
    [
        pytest.param({"Cluster": {}}, None, id="no-log-uri"),
        pytest.param({"Cluster": {"LogUri": "s3://myLogUri/"}}, "myLogUri/", id="has-log-uri"),
    ],
)
def test_get_log_uri(cluster_info, expected_uri):
    emr_client = MagicMock()
    emr_client.describe_cluster.return_value = cluster_info
    assert get_log_uri(cluster=None, emr_client=emr_client, job_flow_id="test_job_flow_id") == expected_uri


class TestEmrLogsLink(BaseAwsLinksTestCase):
    link_class = EmrLogsLink

    def test_extra_link(self, mock_supervisor_comms):
        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.get_message.return_value = XComResult(
                key=self.link_class.key,
                value={
                    "region_name": "eu-west-2",
                    "aws_domain": self.link_class.get_aws_domain("aws"),
                    "log_uri": "myLogUri/",
                    "job_flow_id": "j-8989898989",
                },
            )
        self.assert_extra_link_url(
            expected_url=(
                "https://console.aws.amazon.com/s3/buckets/myLogUri/?region=eu-west-2&prefix=j-8989898989/"
            ),
            region_name="eu-west-2",
            aws_partition="aws",
            log_uri="myLogUri/",
            job_flow_id="j-8989898989",
        )

    @pytest.mark.parametrize(
        "log_url_extra",
        [
            pytest.param({}, id="no-log-uri"),
            pytest.param({"log_uri": None}, id="log-uri-none"),
            pytest.param({"log_uri": ""}, id="log-uri-empty"),
        ],
    )
    def test_missing_log_url(self, log_url_extra: dict):
        self.assert_extra_link_url(expected_url="", **log_url_extra)


@pytest.fixture
def mocked_emr_serverless_hook():
    with mock.patch("airflow.providers.amazon.aws.links.emr.EmrServerlessHook") as m:
        yield m


class TestEmrServerlessLogsLink(BaseAwsLinksTestCase):
    link_class = EmrServerlessLogsLink

    def test_extra_link(self, mocked_emr_serverless_hook, mock_supervisor_comms):
        mocked_client = mocked_emr_serverless_hook.return_value.conn
        mocked_client.get_dashboard_for_job_run.return_value = {"url": "https://example.com/?authToken=1234"}

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.get_message.return_value = XComResult(
                key=self.link_class.key,
                value={
                    "conn_id": "aws-test",
                    "application_id": "app-id",
                    "job_run_id": "job-run-id",
                },
            )

        self.assert_extra_link_url(
            expected_url="https://example.com/logs/SPARK_DRIVER/stdout.gz?authToken=1234",
            conn_id="aws-test",
            application_id="app-id",
            job_run_id="job-run-id",
        )

        mocked_emr_serverless_hook.assert_called_with(
            aws_conn_id="aws-test", config={"retries": {"total_max_attempts": 1}}
        )
        mocked_client.get_dashboard_for_job_run.assert_called_with(
            applicationId="app-id",
            jobRunId="job-run-id",
        )


class TestEmrServerlessDashboardLink(BaseAwsLinksTestCase):
    link_class = EmrServerlessDashboardLink

    def test_extra_link(self, mocked_emr_serverless_hook, mock_supervisor_comms):
        mocked_client = mocked_emr_serverless_hook.return_value.conn
        mocked_client.get_dashboard_for_job_run.return_value = {"url": "https://example.com/?authToken=1234"}

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.get_message.return_value = XComResult(
                key=self.link_class.key,
                value={
                    "conn_id": "aws-test",
                    "application_id": "app-id",
                    "job_run_id": "job-run-id",
                },
            )
        self.assert_extra_link_url(
            expected_url="https://example.com/?authToken=1234",
            conn_id="aws-test",
            application_id="app-id",
            job_run_id="job-run-id",
        )

        mocked_emr_serverless_hook.assert_called_with(
            aws_conn_id="aws-test", config={"retries": {"total_max_attempts": 1}}
        )
        mocked_client.get_dashboard_for_job_run.assert_called_with(
            applicationId="app-id",
            jobRunId="job-run-id",
        )


@pytest.mark.parametrize(
    "dashboard_info, expected_uri",
    [
        pytest.param(
            {"url": "https://example.com/?authToken=first-unique-value"},
            "https://example.com/?authToken=first-unique-value",
            id="first-call",
        ),
        pytest.param(
            {"url": "https://example.com/?authToken=second-unique-value"},
            "https://example.com/?authToken=second-unique-value",
            id="second-call",
        ),
    ],
)
def test_get_serverless_dashboard_url_with_client(mocked_emr_serverless_hook, dashboard_info, expected_uri):
    mocked_client = mocked_emr_serverless_hook.return_value.conn
    mocked_client.get_dashboard_for_job_run.return_value = dashboard_info

    url = get_serverless_dashboard_url(
        emr_serverless_client=mocked_client, application_id="anything", job_run_id="anything"
    )
    assert url
    assert url.geturl() == expected_uri
    mocked_emr_serverless_hook.assert_not_called()
    mocked_client.get_dashboard_for_job_run.assert_called_with(
        applicationId="anything",
        jobRunId="anything",
    )


def test_get_serverless_dashboard_url_with_conn_id(mocked_emr_serverless_hook):
    mocked_client = mocked_emr_serverless_hook.return_value.conn
    mocked_client.get_dashboard_for_job_run.return_value = {
        "url": "https://example.com/?authToken=some-unique-value"
    }

    url = get_serverless_dashboard_url(
        aws_conn_id="aws-test", application_id="anything", job_run_id="anything"
    )
    assert url
    assert url.geturl() == "https://example.com/?authToken=some-unique-value"
    mocked_emr_serverless_hook.assert_called_with(
        aws_conn_id="aws-test", config={"retries": {"total_max_attempts": 1}}
    )
    mocked_client.get_dashboard_for_job_run.assert_called_with(
        applicationId="anything",
        jobRunId="anything",
    )


def test_get_serverless_dashboard_url_parameters():
    with pytest.raises(
        AirflowException, match="Requires either an AWS connection ID or an EMR Serverless Client"
    ):
        get_serverless_dashboard_url(application_id="anything", job_run_id="anything")

    with pytest.raises(
        AirflowException, match="Requires either an AWS connection ID or an EMR Serverless Client"
    ):
        get_serverless_dashboard_url(
            aws_conn_id="a", emr_serverless_client="b", application_id="anything", job_run_id="anything"
        )


class TestEmrServerlessS3LogsLink(BaseAwsLinksTestCase):
    link_class = EmrServerlessS3LogsLink

    def test_extra_link(self, mock_supervisor_comms):
        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.get_message.return_value = XComResult(
                key=self.link_class.key,
                value={
                    "region_name": "us-west-1",
                    "aws_domain": self.link_class.get_aws_domain("aws"),
                    "aws_partition": "aws",
                    "log_uri": "s3://bucket-name/logs/",
                    "application_id": "app-id",
                    "job_run_id": "job-run-id",
                },
            )
        self.assert_extra_link_url(
            expected_url=(
                "https://console.aws.amazon.com/s3/buckets/bucket-name?region=us-west-1&prefix=logs/applications/app-id/jobs/job-run-id/"
            ),
            region_name="us-west-1",
            aws_partition="aws",
            log_uri="s3://bucket-name/logs/",
            application_id="app-id",
            job_run_id="job-run-id",
        )


class TestEmrServerlessCloudWatchLogsLink(BaseAwsLinksTestCase):
    link_class = EmrServerlessCloudWatchLogsLink

    def test_extra_link(self, mock_supervisor_comms):
        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            mock_supervisor_comms.get_message.return_value = XComResult(
                key=self.link_class.key,
                value={
                    "region_name": "us-west-1",
                    "aws_domain": self.link_class.get_aws_domain("aws"),
                    "aws_partition": "aws",
                    "awslogs_group": "/aws/emrs",
                    "stream_prefix": "some-prefix",
                    "application_id": "app-id",
                    "job_run_id": "job-run-id",
                },
            )
        self.assert_extra_link_url(
            expected_url=(
                "https://console.aws.amazon.com/cloudwatch/home?region=us-west-1#logsV2:log-groups/log-group/%2Faws%2Femrs$3FlogStreamNameFilter$3Dsome-prefix"
            ),
            region_name="us-west-1",
            aws_partition="aws",
            awslogs_group="/aws/emrs",
            stream_prefix="some-prefix",
            application_id="app-id",
            job_run_id="job-run-id",
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.amazon.aws.transfers.ftp_to_s3 import FTPToS3Operator

TASK_ID = "test_ftp_to_s3"
BUCKET = "test-s3-bucket"
S3_KEY = "test/test_1_file.csv"
FTP_PATH = "/tmp/remote_path.txt"
AWS_CONN_ID = "aws_default"
FTP_CONN_ID = "ftp_default"
S3_KEY_MULTIPLE = "test/"
FTP_PATH_MULTIPLE = "/tmp/"


class TestFTPToS3Operator:
    def assert_execute(
        self, mock_local_tmp_file, mock_s3_hook_load_file, mock_ftp_hook_retrieve_file, ftp_file, s3_file
    ):
        mock_local_tmp_file_value = mock_local_tmp_file.return_value.__enter__.return_value
        mock_ftp_hook_retrieve_file.assert_called_once_with(
            local_full_path_or_buffer=mock_local_tmp_file_value.name, remote_full_path=ftp_file
        )

        mock_s3_hook_load_file.assert_called_once_with(
            filename=mock_local_tmp_file_value.name,
            key=s3_file,
            bucket_name=BUCKET,
            acl_policy=None,
            encrypt=False,
            gzip=False,
            replace=False,
        )

    @mock.patch("airflow.providers.ftp.hooks.ftp.FTPHook.retrieve_file")
    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.load_file")
    @mock.patch("airflow.providers.amazon.aws.transfers.ftp_to_s3.NamedTemporaryFile")
    def test_execute(self, mock_local_tmp_file, mock_s3_hook_load_file, mock_ftp_hook_retrieve_file):
        operator = FTPToS3Operator(task_id=TASK_ID, s3_bucket=BUCKET, s3_key=S3_KEY, ftp_path=FTP_PATH)
        operator.execute(None)

        self.assert_execute(
            mock_local_tmp_file,
            mock_s3_hook_load_file,
            mock_ftp_hook_retrieve_file,
            ftp_file=operator.ftp_path,
            s3_file=operator.s3_key,
        )

    @mock.patch("airflow.providers.ftp.hooks.ftp.FTPHook.retrieve_file")
    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.load_file")
    @mock.patch("airflow.providers.amazon.aws.transfers.ftp_to_s3.NamedTemporaryFile")
    def test_execute_multiple_files_different_names(
        self, mock_local_tmp_file, mock_s3_hook_load_file, mock_ftp_hook_retrieve_file
    ):
        operator = FTPToS3Operator(
            task_id=TASK_ID,
            s3_bucket=BUCKET,
            s3_key=S3_KEY_MULTIPLE,
            ftp_path=FTP_PATH_MULTIPLE,
            ftp_filenames=["test1.txt"],
            s3_filenames=["test1_s3.txt"],
        )
        operator.execute(None)

        self.assert_execute(
            mock_local_tmp_file,
            mock_s3_hook_load_file,
            mock_ftp_hook_retrieve_file,
            ftp_file=operator.ftp_path + operator.ftp_filenames[0],
            s3_file=operator.s3_key + operator.s3_filenames[0],
        )

    @mock.patch("airflow.providers.ftp.hooks.ftp.FTPHook.retrieve_file")
    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.load_file")
    @mock.patch("airflow.providers.amazon.aws.transfers.ftp_to_s3.NamedTemporaryFile")
    def test_execute_multiple_files_same_names(
        self, mock_local_tmp_file, mock_s3_hook_load_file, mock_ftp_hook_retrieve_file
    ):
        operator = FTPToS3Operator(
            task_id=TASK_ID,
            s3_bucket=BUCKET,
            s3_key=S3_KEY_MULTIPLE,
            ftp_path=FTP_PATH_MULTIPLE,
            ftp_filenames=["test1.txt"],
        )
        operator.execute(None)

        self.assert_execute(
            mock_local_tmp_file,
            mock_s3_hook_load_file,
            mock_ftp_hook_retrieve_file,
            ftp_file=operator.ftp_path + operator.ftp_filenames[0],
            s3_file=operator.s3_key + operator.ftp_filenames[0],
        )

    @mock.patch("airflow.providers.ftp.hooks.ftp.FTPHook.list_directory")
    def test_execute_multiple_files_prefix(
        self,
        mock_ftp_hook_list_directory,
    ):
        operator = FTPToS3Operator(
            task_id=TASK_ID,
            s3_bucket=BUCKET,
            s3_key=S3_KEY_MULTIPLE,
            ftp_path=FTP_PATH_MULTIPLE,
            ftp_filenames="test_prefix",
            s3_filenames="s3_prefix",
        )
        operator.execute(None)

        mock_ftp_hook_list_directory.assert_called_once_with(path=FTP_PATH_MULTIPLE)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import abc
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.configuration import conf
from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.comprehend import ComprehendHook
from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.triggers.comprehend import (
    ComprehendCreateDocumentClassifierCompletedTrigger,
    ComprehendPiiEntitiesDetectionJobCompletedTrigger,
)
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields

if TYPE_CHECKING:
    from airflow.utils.context import Context


class ComprehendBaseSensor(AwsBaseSensor[ComprehendHook]):
    """
    General sensor behavior for Amazon Comprehend.

    Subclasses must implement following methods:
        - ``get_state()``

    Subclasses must set the following fields:
        - ``INTERMEDIATE_STATES``
        - ``FAILURE_STATES``
        - ``SUCCESS_STATES``
        - ``FAILURE_MESSAGE``

    :param deferrable: If True, the sensor will operate in deferrable mode. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    """

    aws_hook_class = ComprehendHook

    INTERMEDIATE_STATES: tuple[str, ...] = ()
    FAILURE_STATES: tuple[str, ...] = ()
    SUCCESS_STATES: tuple[str, ...] = ()
    FAILURE_MESSAGE = ""

    ui_color = "#66c3ff"

    def __init__(
        self,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.deferrable = deferrable

    def poke(self, context: Context, **kwargs) -> bool:
        state = self.get_state()
        if state in self.FAILURE_STATES:
            raise AirflowException(self.FAILURE_MESSAGE)

        return state not in self.INTERMEDIATE_STATES

    @abc.abstractmethod
    def get_state(self) -> str:
        """Implement in subclasses."""


class ComprehendStartPiiEntitiesDetectionJobCompletedSensor(ComprehendBaseSensor):
    """
    Poll the state of the pii entities detection job until it reaches a completed state; fails if the job fails.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:ComprehendStartPiiEntitiesDetectionJobCompletedSensor`

    :param job_id: The id of the Comprehend pii entities detection job.

    :param deferrable: If True, the sensor will operate in deferrable mode. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    :param poke_interval: Polling period in seconds to check for the status of the job. (default: 120)
    :param max_retries: Number of times before returning the current state. (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    INTERMEDIATE_STATES: tuple[str, ...] = ("IN_PROGRESS",)
    FAILURE_STATES: tuple[str, ...] = ("FAILED", "STOP_REQUESTED", "STOPPED")
    SUCCESS_STATES: tuple[str, ...] = ("COMPLETED",)
    FAILURE_MESSAGE = "Comprehend start pii entities detection job sensor failed."

    template_fields: Sequence[str] = aws_template_fields("job_id")

    def __init__(
        self,
        *,
        job_id: str,
        max_retries: int = 75,
        poke_interval: int = 120,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.job_id = job_id
        self.max_retries = max_retries
        self.poke_interval = poke_interval

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=ComprehendPiiEntitiesDetectionJobCompletedTrigger(
                    job_id=self.job_id,
                    waiter_delay=int(self.poke_interval),
                    waiter_max_attempts=self.max_retries,
                    aws_conn_id=self.aws_conn_id,
                ),
                method_name="poke",
            )
        else:
            super().execute(context=context)

    def get_state(self) -> str:
        return self.hook.conn.describe_pii_entities_detection_job(JobId=self.job_id)[
            "PiiEntitiesDetectionJobProperties"
        ]["JobStatus"]


class ComprehendCreateDocumentClassifierCompletedSensor(AwsBaseSensor[ComprehendHook]):
    """
    Poll the state of the document classifier until it reaches a completed state; fails if the job fails.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:ComprehendCreateDocumentClassifierCompletedSensor`

    :param document_classifier_arn: The arn of the Comprehend document classifier.
    :param fail_on_warnings: If set to True, the document classifier training job will throw an error when the
        status is TRAINED_WITH_WARNING. (default False)

    :param deferrable: If True, the sensor will operate in deferrable mode. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    :param poke_interval: Polling period in seconds to check for the status of the job. (default: 120)
    :param max_retries: Number of times before returning the current state. (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = ComprehendHook

    INTERMEDIATE_STATES: tuple[str, ...] = (
        "SUBMITTED",
        "TRAINING",
    )
    FAILURE_STATES: tuple[str, ...] = (
        "DELETING",
        "STOP_REQUESTED",
        "STOPPED",
        "IN_ERROR",
    )
    SUCCESS_STATES: tuple[str, ...] = ("TRAINED", "TRAINED_WITH_WARNING")
    FAILURE_MESSAGE = "Comprehend document classifier failed."

    template_fields: Sequence[str] = aws_template_fields("document_classifier_arn")

    def __init__(
        self,
        *,
        document_classifier_arn: str,
        fail_on_warnings: bool = False,
        max_retries: int = 75,
        poke_interval: int = 120,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        aws_conn_id: str | None = "aws_default",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.document_classifier_arn = document_classifier_arn
        self.fail_on_warnings = fail_on_warnings
        self.max_retries = max_retries
        self.poke_interval = poke_interval
        self.deferrable = deferrable
        self.aws_conn_id = aws_conn_id

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=ComprehendCreateDocumentClassifierCompletedTrigger(
                    document_classifier_arn=self.document_classifier_arn,
                    waiter_delay=int(self.poke_interval),
                    waiter_max_attempts=self.max_retries,
                    aws_conn_id=self.aws_conn_id,
                ),
                method_name="poke",
            )
        else:
            super().execute(context=context)

    def poke(self, context: Context, **kwargs) -> bool:
        status = self.hook.conn.describe_document_classifier(
            DocumentClassifierArn=self.document_classifier_arn
        )["DocumentClassifierProperties"]["Status"]

        self.log.info(
            "Poking for AWS Comprehend document classifier arn: %s status: %s",
            self.document_classifier_arn,
            status,
        )

        if status in self.FAILURE_STATES:
            raise AirflowException(self.FAILURE_MESSAGE)

        if status in self.SUCCESS_STATES:
            self.hook.validate_document_classifier_training_status(
                document_classifier_arn=self.document_classifier_arn, fail_on_warnings=self.fail_on_warnings
            )

            self.log.info("Comprehend document classifier `%s` complete.", self.document_classifier_arn)

            return True

        return False



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock, call, patch

import pytest
from sagemaker_studio.models.execution import ExecutionClient

from airflow.exceptions import AirflowException
from airflow.models import TaskInstance
from airflow.providers.amazon.aws.hooks.sagemaker_unified_studio import (
    SageMakerNotebookHook,
)
from airflow.utils.session import create_session

pytestmark = pytest.mark.db_test


class TestSageMakerNotebookHook:
    @pytest.fixture(autouse=True)
    def setup(self):
        with patch(
            "airflow.providers.amazon.aws.hooks.sagemaker_unified_studio.SageMakerStudioAPI",
            autospec=True,
        ) as mock_sdk:
            self.execution_name = "test-execution"
            self.waiter_delay = 10
            sdk_instance = mock_sdk.return_value
            sdk_instance.execution_client = MagicMock(spec=ExecutionClient)
            sdk_instance.execution_client.start_execution.return_value = {
                "execution_id": "execution_id",
                "execution_name": "execution_name",
            }
            self.hook = SageMakerNotebookHook(
                input_config={
                    "input_path": "test-data/notebook/test_notebook.ipynb",
                    "input_params": {"key": "value"},
                },
                output_config={"output_formats": ["NOTEBOOK"]},
                execution_name=self.execution_name,
                waiter_delay=self.waiter_delay,
                compute={"instance_type": "ml.c4.2xlarge"},
            )

            self.hook._sagemaker_studio = mock_sdk
            self.files = [
                {"display_name": "file1.txt", "url": "http://example.com/file1.txt"},
                {"display_name": "file2.txt", "url": "http://example.com/file2.txt"},
            ]
            self.context = {
                "ti": MagicMock(spec=TaskInstance),
            }
            self.s3Path = "S3Path"
            yield

    def test_format_input_config(self):
        expected_config = {
            "notebook_config": {
                "input_path": "test-data/notebook/test_notebook.ipynb",
                "input_parameters": {"key": "value"},
            }
        }

        config = self.hook._format_start_execution_input_config()
        assert config == expected_config

    def test_format_output_config(self):
        expected_config = {
            "notebook_config": {
                "output_formats": ["NOTEBOOK"],
            }
        }

        config = self.hook._format_start_execution_output_config()
        assert config == expected_config

    def test_format_output_config_default(self):
        no_output_config_hook = SageMakerNotebookHook(
            input_config={
                "input_path": "test-data/notebook/test_notebook.ipynb",
                "input_params": {"key": "value"},
            },
            execution_name=self.execution_name,
            waiter_delay=self.waiter_delay,
        )

        no_output_config_hook._sagemaker_studio = self.hook._sagemaker_studio
        expected_config = {"notebook_config": {"output_formats": ["NOTEBOOK"]}}

        config = no_output_config_hook._format_start_execution_output_config()
        assert config == expected_config

    def test_start_notebook_execution(self):
        self.hook._sagemaker_studio = MagicMock()
        self.hook._sagemaker_studio.execution_client = MagicMock(spec=ExecutionClient)

        self.hook._sagemaker_studio.execution_client.start_execution.return_value = {"executionId": "123456"}
        result = self.hook.start_notebook_execution()
        assert result == {"executionId": "123456"}
        self.hook._sagemaker_studio.execution_client.start_execution.assert_called_once()

    @patch("time.sleep", return_value=None)  # To avoid actual sleep during tests
    def test_wait_for_execution_completion(self, mock_sleep):
        execution_id = "123456"
        self.hook._sagemaker_studio = MagicMock()
        self.hook._sagemaker_studio.execution_client = MagicMock(spec=ExecutionClient)
        self.hook._sagemaker_studio.execution_client.get_execution.return_value = {"status": "COMPLETED"}

        result = self.hook.wait_for_execution_completion(execution_id, {})
        assert result == {"Status": "COMPLETED", "ExecutionId": execution_id}
        self.hook._sagemaker_studio.execution_client.get_execution.assert_called()
        mock_sleep.assert_called_once()

    @patch("time.sleep", return_value=None)
    def test_wait_for_execution_completion_failed(self, mock_sleep):
        execution_id = "123456"
        self.hook._sagemaker_studio = MagicMock()
        self.hook._sagemaker_studio.execution_client = MagicMock(spec=ExecutionClient)
        self.hook._sagemaker_studio.execution_client.get_execution.return_value = {
            "status": "FAILED",
            "error_details": {"error_message": "Execution failed"},
        }

        with pytest.raises(AirflowException, match="Execution failed"):
            self.hook.wait_for_execution_completion(execution_id, self.context)

    def test_handle_in_progress_state(self):
        execution_id = "123456"
        states = ["IN_PROGRESS", "STOPPING"]

        for status in states:
            result = self.hook._handle_state(execution_id, status, None)
            assert result is None

    def test_handle_finished_state(self):
        execution_id = "123456"
        states = ["COMPLETED"]

        for status in states:
            result = self.hook._handle_state(execution_id, status, None)
            assert result == {"Status": status, "ExecutionId": execution_id}

    def test_handle_failed_state(self):
        execution_id = "123456"
        status = "FAILED"
        error_message = "Execution failed"
        with pytest.raises(AirflowException, match=error_message):
            self.hook._handle_state(execution_id, status, error_message)

        status = "STOPPED"
        error_message = ""
        with pytest.raises(AirflowException, match=f"Exiting Execution {execution_id} State: {status}"):
            self.hook._handle_state(execution_id, status, error_message)

    def test_handle_unexpected_state(self):
        execution_id = "123456"
        status = "PENDING"
        error_message = f"Exiting Execution {execution_id} State: {status}"
        with pytest.raises(AirflowException, match=error_message):
            self.hook._handle_state(execution_id, status, error_message)

    @patch(
        "airflow.providers.amazon.aws.hooks.sagemaker_unified_studio.SageMakerNotebookHook._set_xcom_files"
    )
    def test_set_xcom_files(self, mock_set_xcom_files):
        with create_session():
            self.hook._set_xcom_files(self.files, self.context)
        expected_call = call(self.files, self.context)
        mock_set_xcom_files.assert_called_once_with(*expected_call.args, **expected_call.kwargs)

    def test_set_xcom_files_negative_missing_context(self):
        with pytest.raises(AirflowException, match="context is required"):
            self.hook._set_xcom_files(self.files, {})

    @patch(
        "airflow.providers.amazon.aws.hooks.sagemaker_unified_studio.SageMakerNotebookHook._set_xcom_s3_path"
    )
    def test_set_xcom_s3_path(self, mock_set_xcom_s3_path):
        with create_session():
            self.hook._set_xcom_s3_path(self.s3Path, self.context)
        expected_call = call(self.s3Path, self.context)
        mock_set_xcom_s3_path.assert_called_once_with(*expected_call.args, **expected_call.kwargs)

    def test_set_xcom_s3_path_negative_missing_context(self):
        with pytest.raises(AirflowException, match="context is required"):
            self.hook._set_xcom_s3_path(self.s3Path, {})



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

import boto3

from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.sagemaker import (
    SageMakerStartPipelineOperator,
    SageMakerStopPipelineOperator,
)
from airflow.providers.amazon.aws.sensors.sagemaker import (
    SageMakerPipelineSensor,
)
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.example_sagemaker import delete_experiments
from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

DAG_ID = "example_sagemaker_pipeline"

# Externally fetched variables:
ROLE_ARN_KEY = "ROLE_ARN"

sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()


@task
def create_pipeline(name: str, role_arn: str):
    # Json definition for a dummy pipeline of 30 chained "conditional step" checking that 3 < 6
    # Each step takes roughly 1 second to execute, so the pipeline runtimes is ~30 seconds, which should be
    # enough to test stopping and awaiting without race conditions.
    # Built using sagemaker sdk, and using json.loads(pipeline.definition())
    pipeline_json_definition = """{"Version": "2020-12-01", "Metadata": {}, "Parameters": [], "PipelineExperimentConfig": {"ExperimentName": {"Get": "Execution.PipelineName"}, "TrialName": {"Get": "Execution.PipelineExecutionId"}}, "Steps": [{"Name": "DummyCond29", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond28", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond27", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond26", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond25", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond24", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond23", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond22", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond21", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond20", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond19", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond18", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond17", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond16", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond15", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond14", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond13", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond12", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond11", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond10", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond9", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond8", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond7", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond6", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond5", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond4", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond3", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond2", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond1", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond0", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [{"Name": "DummyCond", "Type": "Condition", "Arguments": {"Conditions": [{"Type": "LessThanOrEqualTo", "LeftValue": 3.0, "RightValue": 6.0}], "IfSteps": [], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}], "ElseSteps": []}}]}"""
    sgmk_client = boto3.client("sagemaker")
    sgmk_client.create_pipeline(
        PipelineName=name, PipelineDefinition=pipeline_json_definition, RoleArn=role_arn
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_pipeline(name: str):
    sgmk_client = boto3.client("sagemaker")
    sgmk_client.delete_pipeline(PipelineName=name)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]

    pipeline_name = f"{env_id}-pipeline"

    create_pipeline = create_pipeline(pipeline_name, test_context[ROLE_ARN_KEY])

    # [START howto_operator_sagemaker_start_pipeline]
    start_pipeline1 = SageMakerStartPipelineOperator(
        task_id="start_pipeline1",
        pipeline_name=pipeline_name,
    )
    # [END howto_operator_sagemaker_start_pipeline]

    # [START howto_operator_sagemaker_stop_pipeline]
    stop_pipeline1 = SageMakerStopPipelineOperator(
        task_id="stop_pipeline1",
        pipeline_exec_arn=start_pipeline1.output,
    )
    # [END howto_operator_sagemaker_stop_pipeline]

    start_pipeline2 = SageMakerStartPipelineOperator(
        task_id="start_pipeline2",
        pipeline_name=pipeline_name,
    )

    # [START howto_sensor_sagemaker_pipeline]
    await_pipeline2 = SageMakerPipelineSensor(
        task_id="await_pipeline2",
        pipeline_exec_arn=start_pipeline2.output,
    )
    # [END howto_sensor_sagemaker_pipeline]
    await_pipeline2.poke_interval = 10

    chain(
        # TEST SETUP
        test_context,
        create_pipeline,  # type: ignore[arg-type]
        # TEST BODY
        start_pipeline1,
        start_pipeline2,
        stop_pipeline1,
        await_pipeline2,
        # TEST TEARDOWN
        delete_pipeline(pipeline_name),
        delete_experiments([pipeline_name]),
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.exceptions import AirflowException, TaskDeferred
from airflow.providers.amazon.aws.hooks.batch_client import BatchClientHook
from airflow.providers.amazon.aws.sensors.batch import (
    BatchComputeEnvironmentSensor,
    BatchJobQueueSensor,
    BatchSensor,
)
from airflow.providers.amazon.aws.triggers.batch import BatchJobTrigger

TASK_ID = "batch_job_sensor"
JOB_ID = "8222a1c2-b246-4e19-b1b8-0039bb4407c0"
AWS_REGION = "eu-west-1"
ENVIRONMENT_NAME = "environment_name"
JOB_QUEUE = "job_queue"


@pytest.fixture(scope="module")
def batch_sensor() -> BatchSensor:
    return BatchSensor(
        task_id="batch_job_sensor",
        job_id=JOB_ID,
    )


@pytest.fixture(scope="module")
def deferrable_batch_sensor() -> BatchSensor:
    return BatchSensor(task_id="task", job_id=JOB_ID, region_name=AWS_REGION, deferrable=True)


class TestBatchSensor:
    @mock.patch.object(BatchClientHook, "get_job_description")
    def test_poke_on_success_state(self, mock_get_job_description, batch_sensor: BatchSensor):
        mock_get_job_description.return_value = {"status": "SUCCEEDED"}
        assert batch_sensor.poke({}) is True
        mock_get_job_description.assert_called_once_with(JOB_ID)

    @mock.patch.object(BatchClientHook, "get_job_description")
    def test_poke_on_failure_state(self, mock_get_job_description, batch_sensor: BatchSensor):
        mock_get_job_description.return_value = {"status": "FAILED"}
        with pytest.raises(AirflowException, match="Batch sensor failed. AWS Batch job status: FAILED"):
            batch_sensor.poke({})

        mock_get_job_description.assert_called_once_with(JOB_ID)

    @mock.patch.object(BatchClientHook, "get_job_description")
    def test_poke_on_invalid_state(self, mock_get_job_description, batch_sensor: BatchSensor):
        mock_get_job_description.return_value = {"status": "INVALID"}
        with pytest.raises(AirflowException, match="Batch sensor failed. AWS Batch job status: INVALID"):
            batch_sensor.poke({})

        mock_get_job_description.assert_called_once_with(JOB_ID)

    @pytest.mark.parametrize("job_status", ["SUBMITTED", "PENDING", "RUNNABLE", "STARTING", "RUNNING"])
    @mock.patch.object(BatchClientHook, "get_job_description")
    def test_poke_on_intermediate_state(
        self, mock_get_job_description, job_status, batch_sensor: BatchSensor
    ):
        print(job_status)
        mock_get_job_description.return_value = {"status": job_status}
        assert batch_sensor.poke({}) is False
        mock_get_job_description.assert_called_once_with(JOB_ID)

    def test_execute_in_deferrable_mode(self, deferrable_batch_sensor: BatchSensor):
        """
        Asserts that a task is deferred and a BatchSensorTrigger will be fired
        when the BatchSensor is executed in deferrable mode.
        """

        with pytest.raises(TaskDeferred) as exc:
            deferrable_batch_sensor.execute({})
        assert isinstance(exc.value.trigger, BatchJobTrigger), "Trigger is not a BatchJobTrigger"

    def test_execute_failure_in_deferrable_mode(self, deferrable_batch_sensor: BatchSensor):
        """Tests that an AirflowException is raised in case of error event"""

        with pytest.raises(AirflowException):
            deferrable_batch_sensor.execute_complete(context={}, event={"status": "failure"})

    @pytest.mark.parametrize(
        "state",
        (
            BatchClientHook.FAILURE_STATE,
            "unknown_state",
        ),
    )
    @mock.patch.object(BatchClientHook, "get_job_description")
    def test_fail_poke(self, mock_get_job_description, state):
        mock_get_job_description.return_value = {"status": state}
        batch_sensor = BatchSensor(task_id="batch_job_sensor", job_id=JOB_ID)
        with pytest.raises(AirflowException):
            batch_sensor.poke({})


@pytest.fixture(scope="module")
def batch_compute_environment_sensor() -> BatchComputeEnvironmentSensor:
    return BatchComputeEnvironmentSensor(
        task_id="test_batch_compute_environment_sensor",
        compute_environment=ENVIRONMENT_NAME,
    )


class TestBatchComputeEnvironmentSensor:
    @mock.patch.object(BatchClientHook, "client")
    def test_poke_no_environment(
        self, mock_batch_client, batch_compute_environment_sensor: BatchComputeEnvironmentSensor
    ):
        mock_batch_client.describe_compute_environments.return_value = {"computeEnvironments": []}
        with pytest.raises(AirflowException) as ctx:
            batch_compute_environment_sensor.poke({})
        mock_batch_client.describe_compute_environments.assert_called_once_with(
            computeEnvironments=[ENVIRONMENT_NAME],
        )
        assert "not found" in str(ctx.value)

    @mock.patch.object(BatchClientHook, "client")
    def test_poke_valid(
        self, mock_batch_client, batch_compute_environment_sensor: BatchComputeEnvironmentSensor
    ):
        mock_batch_client.describe_compute_environments.return_value = {
            "computeEnvironments": [{"status": "VALID"}]
        }
        assert batch_compute_environment_sensor.poke({}) is True
        mock_batch_client.describe_compute_environments.assert_called_once_with(
            computeEnvironments=[ENVIRONMENT_NAME],
        )

    @mock.patch.object(BatchClientHook, "client")
    def test_poke_running(
        self, mock_batch_client, batch_compute_environment_sensor: BatchComputeEnvironmentSensor
    ):
        mock_batch_client.describe_compute_environments.return_value = {
            "computeEnvironments": [
                {
                    "status": "CREATING",
                }
            ]
        }
        assert batch_compute_environment_sensor.poke({}) is False
        mock_batch_client.describe_compute_environments.assert_called_once_with(
            computeEnvironments=[ENVIRONMENT_NAME],
        )

    @mock.patch.object(BatchClientHook, "client")
    def test_poke_invalid(
        self, mock_batch_client, batch_compute_environment_sensor: BatchComputeEnvironmentSensor
    ):
        mock_batch_client.describe_compute_environments.return_value = {
            "computeEnvironments": [
                {
                    "status": "INVALID",
                }
            ]
        }
        with pytest.raises(AirflowException) as ctx:
            batch_compute_environment_sensor.poke({})
        mock_batch_client.describe_compute_environments.assert_called_once_with(
            computeEnvironments=[ENVIRONMENT_NAME],
        )
        assert "AWS Batch compute environment failed" in str(ctx.value)

    @pytest.mark.parametrize(
        "compute_env, error_message",
        (
            (
                [{"status": "unknown_status"}],
                "AWS Batch compute environment failed. AWS Batch compute environment status:",
            ),
            ([], "AWS Batch compute environment"),
        ),
    )
    @mock.patch.object(BatchClientHook, "client")
    def test_fail_poke(
        self,
        mock_batch_client,
        batch_compute_environment_sensor: BatchComputeEnvironmentSensor,
        compute_env,
        error_message,
    ):
        mock_batch_client.describe_compute_environments.return_value = {"computeEnvironments": compute_env}
        with pytest.raises(AirflowException, match=error_message):
            batch_compute_environment_sensor.poke({})


@pytest.fixture(scope="module")
def batch_job_queue_sensor() -> BatchJobQueueSensor:
    return BatchJobQueueSensor(
        task_id="test_batch_job_queue_sensor",
        job_queue=JOB_QUEUE,
    )


class TestBatchJobQueueSensor:
    @mock.patch.object(BatchClientHook, "client")
    def test_poke_no_queue(self, mock_batch_client, batch_job_queue_sensor: BatchJobQueueSensor):
        mock_batch_client.describe_job_queues.return_value = {"jobQueues": []}
        with pytest.raises(AirflowException) as ctx:
            batch_job_queue_sensor.poke({})
        mock_batch_client.describe_job_queues.assert_called_once_with(
            jobQueues=[JOB_QUEUE],
        )
        assert "not found" in str(ctx.value)

    @mock.patch.object(BatchClientHook, "client")
    def test_poke_no_queue_with_treat_non_existing_as_deleted(
        self, mock_batch_client, batch_job_queue_sensor: BatchJobQueueSensor
    ):
        batch_job_queue_sensor.treat_non_existing_as_deleted = True
        mock_batch_client.describe_job_queues.return_value = {"jobQueues": []}
        assert batch_job_queue_sensor.poke({}) is True
        mock_batch_client.describe_job_queues.assert_called_once_with(
            jobQueues=[JOB_QUEUE],
        )

    @mock.patch.object(BatchClientHook, "client")
    def test_poke_valid(self, mock_batch_client, batch_job_queue_sensor: BatchJobQueueSensor):
        mock_batch_client.describe_job_queues.return_value = {"jobQueues": [{"status": "VALID"}]}
        assert batch_job_queue_sensor.poke({}) is True
        mock_batch_client.describe_job_queues.assert_called_once_with(
            jobQueues=[JOB_QUEUE],
        )

    @mock.patch.object(BatchClientHook, "client")
    def test_poke_running(self, mock_batch_client, batch_job_queue_sensor: BatchJobQueueSensor):
        mock_batch_client.describe_job_queues.return_value = {
            "jobQueues": [
                {
                    "status": "CREATING",
                }
            ]
        }
        assert batch_job_queue_sensor.poke({}) is False
        mock_batch_client.describe_job_queues.assert_called_once_with(
            jobQueues=[JOB_QUEUE],
        )

    @mock.patch.object(BatchClientHook, "client")
    def test_poke_invalid(self, mock_batch_client, batch_job_queue_sensor: BatchJobQueueSensor):
        mock_batch_client.describe_job_queues.return_value = {
            "jobQueues": [
                {
                    "status": "INVALID",
                }
            ]
        }
        with pytest.raises(AirflowException) as ctx:
            batch_job_queue_sensor.poke({})
        mock_batch_client.describe_job_queues.assert_called_once_with(
            jobQueues=[JOB_QUEUE],
        )
        assert "AWS Batch job queue failed" in str(ctx.value)

    @pytest.mark.parametrize("job_queue", ([], [{"status": "UNKNOWN_STATUS"}]))
    @mock.patch.object(BatchClientHook, "client")
    def test_fail_poke(
        self,
        mock_batch_client,
        batch_job_queue_sensor: BatchJobQueueSensor,
        job_queue,
    ):
        mock_batch_client.describe_job_queues.return_value = {"jobQueues": job_queue}
        batch_job_queue_sensor.treat_non_existing_as_deleted = False
        message = "AWS Batch job queue"
        with pytest.raises(AirflowException, match=message):
            batch_job_queue_sensor.poke({})



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from botocore.exceptions import ClientError
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.sqs import SqsHook
from airflow.providers.amazon.aws.operators.sqs import SqsPublishOperator

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

REGION_NAME = "eu-west-1"
QUEUE_NAME = "test-queue"
QUEUE_URL = f"https://{QUEUE_NAME}"
FIFO_QUEUE_NAME = "test-queue.fifo"
FIFO_QUEUE_URL = f"https://{FIFO_QUEUE_NAME}"


@pytest.fixture
def mocked_context():
    return mock.MagicMock(name="FakeContext")


class TestSqsPublishOperator:
    @pytest.fixture(autouse=True)
    def _setup_test_cases(self):
        self.default_op_kwargs = {
            "task_id": "test_task",
            "message_content": "hello",
            "aws_conn_id": None,
            "region_name": REGION_NAME,
        }
        self.sqs_client = SqsHook(aws_conn_id=None, region_name=REGION_NAME).conn

    def test_init(self):
        self.default_op_kwargs.pop("aws_conn_id", None)
        self.default_op_kwargs.pop("region_name", None)

        op = SqsPublishOperator(sqs_queue=QUEUE_NAME, **self.default_op_kwargs)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

        op = SqsPublishOperator(
            sqs_queue=FIFO_QUEUE_NAME,
            **self.default_op_kwargs,
            aws_conn_id=None,
            region_name=REGION_NAME,
            verify="/spam/egg.pem",
            botocore_config={"read_timeout": 42},
        )
        assert op.hook.aws_conn_id is None
        assert op.hook._region_name == REGION_NAME
        assert op.hook._verify == "/spam/egg.pem"
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

    @mock_aws
    def test_execute_success(self, mocked_context):
        self.sqs_client.create_queue(QueueName=QUEUE_NAME)

        # Send SQS Message
        op = SqsPublishOperator(**self.default_op_kwargs, sqs_queue=QUEUE_NAME)
        result = op.execute(mocked_context)
        assert "MD5OfMessageBody" in result
        assert "MessageId" in result

        # Validate message through moto
        message = self.sqs_client.receive_message(QueueUrl=QUEUE_URL)
        assert len(message["Messages"]) == 1
        assert message["Messages"][0]["MessageId"] == result["MessageId"]
        assert message["Messages"][0]["Body"] == "hello"

    @mock_aws
    def test_execute_failure_fifo_queue(self, mocked_context):
        self.sqs_client.create_queue(QueueName=FIFO_QUEUE_NAME, Attributes={"FifoQueue": "true"})

        op = SqsPublishOperator(**self.default_op_kwargs, sqs_queue=FIFO_QUEUE_NAME)
        error_message = (
            r"An error occurred \(MissingParameter\) when calling the SendMessage operation: "
            r"The request must contain the parameter MessageGroupId."
        )
        with pytest.raises(ClientError, match=error_message):
            op.execute(mocked_context)

    @mock_aws
    def test_deduplication_failure(self, mocked_context):
        self.sqs_client.create_queue(
            QueueName=FIFO_QUEUE_NAME, Attributes={"FifoQueue": "true", "ContentBasedDeduplication": "false"}
        )

        op = SqsPublishOperator(**self.default_op_kwargs, sqs_queue=FIFO_QUEUE_NAME, message_group_id="abc")
        error_message = (
            r"An error occurred \(InvalidParameterValue\) when calling the SendMessage operation: "
            r"The queue should either have ContentBasedDeduplication enabled or MessageDeduplicationId provided explicitly"
        )
        with pytest.raises(ClientError, match=error_message):
            op.execute(mocked_context)

    @mock_aws
    def test_execute_success_fifo_queue(self, mocked_context):
        self.sqs_client.create_queue(
            QueueName=FIFO_QUEUE_NAME, Attributes={"FifoQueue": "true", "ContentBasedDeduplication": "true"}
        )

        # Send SQS Message into the FIFO Queue
        op = SqsPublishOperator(
            **self.default_op_kwargs,
            sqs_queue=FIFO_QUEUE_NAME,
            message_group_id="abc",
            message_deduplication_id="abc",
        )
        result = op.execute(mocked_context)
        assert "MD5OfMessageBody" in result
        assert "MessageId" in result

        # Validate message through moto
        message = self.sqs_client.receive_message(
            QueueUrl=FIFO_QUEUE_URL, AttributeNames=["MessageGroupId", "MessageDeduplicationId"]
        )
        assert len(message["Messages"]) == 1
        assert message["Messages"][0]["MessageId"] == result["MessageId"]
        assert message["Messages"][0]["Body"] == "hello"
        assert message["Messages"][0]["Attributes"]["MessageGroupId"] == "abc"
        assert message["Messages"][0]["Attributes"]["MessageDeduplicationId"] == "abc"

    def test_template_fields(self):
        operator = SqsPublishOperator(
            **self.default_op_kwargs,
            sqs_queue=FIFO_QUEUE_NAME,
            message_group_id="abc",
            message_deduplication_id="abc",
        )
        validate_template_fields(operator)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.sensors.sagemaker_unified_studio import (
    SageMakerNotebookSensor,
)
from airflow.utils.context import Context


class TestSageMakerNotebookSensor:
    def test_init(self):
        # Test the initialization of the sensor
        sensor = SageMakerNotebookSensor(
            task_id="test_task",
            execution_id="test_execution_id",
            execution_name="test_execution_name",
        )
        assert sensor.execution_id == "test_execution_id"
        assert sensor.execution_name == "test_execution_name"
        assert sensor.success_state == ["COMPLETED"]
        assert sensor.in_progress_states == ["PENDING", "RUNNING"]

    @patch("airflow.providers.amazon.aws.sensors.sagemaker_unified_studio.SageMakerNotebookHook")
    def test_poke_success_state(self, mock_notebook_hook):
        mock_hook_instance = mock_notebook_hook.return_value
        mock_hook_instance.get_execution_status.return_value = "COMPLETED"

        sensor = SageMakerNotebookSensor(
            task_id="test_task",
            execution_id="test_execution_id",
            execution_name="test_execution_name",
        )

        # Test the poke method
        result = sensor.poke()
        assert result is True
        mock_hook_instance.get_execution_status.assert_called_once_with(execution_id="test_execution_id")

    @patch("airflow.providers.amazon.aws.sensors.sagemaker_unified_studio.SageMakerNotebookHook")
    def test_poke_failure_state(self, mock_notebook_hook):
        mock_hook_instance = mock_notebook_hook.return_value
        mock_hook_instance.get_execution_status.return_value = "FAILED"

        sensor = SageMakerNotebookSensor(
            task_id="test_task",
            execution_id="test_execution_id",
            execution_name="test_execution_name",
        )

        # Test the poke method and assert exception
        with pytest.raises(AirflowException, match="Exiting Execution test_execution_id State: FAILED"):
            sensor.poke()

        mock_hook_instance.get_execution_status.assert_called_once_with(execution_id="test_execution_id")

    @patch("airflow.providers.amazon.aws.sensors.sagemaker_unified_studio.SageMakerNotebookHook")
    def test_poke_in_progress_state(self, mock_notebook_hook):
        mock_hook_instance = mock_notebook_hook.return_value
        mock_hook_instance.get_execution_status.return_value = "RUNNING"

        sensor = SageMakerNotebookSensor(
            task_id="test_task",
            execution_id="test_execution_id",
            execution_name="test_execution_name",
        )

        # Test the poke method
        result = sensor.poke()
        assert result is False
        mock_hook_instance.get_execution_status.assert_called_once_with(execution_id="test_execution_id")

    @patch.object(SageMakerNotebookSensor, "poke", return_value=True)
    def test_execute_calls_poke(self, mock_poke):
        # Create the sensor
        sensor = SageMakerNotebookSensor(
            task_id="test_task",
            execution_id="test_execution_id",
            execution_name="test_execution_name",
        )

        context = MagicMock(spec=Context)
        sensor.execute(context=context)

        # Assert that the poke method was called
        mock_poke.assert_called_once_with(context)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow import settings
from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.connection import Connection
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.s3 import S3CreateBucketOperator, S3DeleteBucketOperator
from airflow.providers.amazon.aws.transfers.http_to_s3 import HttpToS3Operator
from airflow.providers.standard.operators.bash import BashOperator
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import SystemTestContextBuilder

sys_test_context_task = SystemTestContextBuilder().build()

DAG_ID = "example_http_to_s3"

cmd = """
#!/bin/bash

echo 'foo' > /tmp/test_file

cd /tmp

nohup python3 -m http.server 8083 > /dev/null 2>&1 &

echo $!
sleep 2
exit 0
"""


@task
def create_connection(conn_id_name: str):
    conn = Connection(
        conn_id=conn_id_name,
        conn_type="http",
        host="localhost",
        port=8083,
    )
    session = settings.Session()
    session.add(conn)
    session.commit()


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example"],
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]

    conn_id_name = f"{env_id}-conn-id"
    s3_bucket = f"{env_id}-http-to-s3-bucket"
    s3_key = f"{env_id}-http-to-s3-key"

    create_s3_bucket = S3CreateBucketOperator(task_id="create_s3_bucket", bucket_name=s3_bucket)

    set_up_connection = create_connection(conn_id_name)

    start_server = BashOperator(bash_command=cmd, task_id="start_server")

    # [START howto_transfer_http_to_s3]
    http_to_s3_task = HttpToS3Operator(
        task_id="http_to_s3_task",
        http_conn_id=conn_id_name,
        endpoint="/test_file",
        s3_bucket=s3_bucket,
        s3_key=s3_key,
        replace=True,
    )
    # [END howto_transfer_http_to_s3]

    stop_server = BashOperator(
        task_id="stop_simple_http_server",
        bash_command='kill {{ti.xcom_pull(task_ids="start_server")}}',
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_s3_bucket = S3DeleteBucketOperator(
        task_id="delete_s3_bucket",
        bucket_name=s3_bucket,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    chain(
        # TEST SETUP
        test_context,
        create_s3_bucket,
        set_up_connection,
        start_server,
        # TEST BODY
        http_to_s3_task,
        # TEST TEARDOWN
        stop_server,
        delete_s3_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.sns import SnsHook


@mock_aws
class TestSnsHook:
    def test_get_conn_returns_a_boto3_connection(self):
        hook = SnsHook(aws_conn_id="aws_default")
        assert hook.get_conn() is not None

    def test_publish_to_target_with_subject(self):
        hook = SnsHook(aws_conn_id="aws_default")

        message = "Hello world"
        topic_name = "test-topic"
        subject = "test-subject"
        target = hook.get_conn().create_topic(Name=topic_name).get("TopicArn")

        response = hook.publish_to_target(target, message, subject)

        assert "MessageId" in response

    def test_publish_to_target_with_attributes(self):
        hook = SnsHook(aws_conn_id="aws_default")

        message = "Hello world"
        topic_name = "test-topic"
        target = hook.get_conn().create_topic(Name=topic_name).get("TopicArn")

        response = hook.publish_to_target(
            target,
            message,
            message_attributes={
                "test-string": "string-value",
                "test-number": 123456,
                "test-array": ["first", "second", "third"],
                "test-binary": b"binary-value",
            },
        )

        assert "MessageId" in response

    def test_publish_to_target_plain(self):
        hook = SnsHook(aws_conn_id="aws_default")

        message = "Hello world"
        topic_name = "test-topic"
        target = hook.get_conn().create_topic(Name=topic_name).get("TopicArn")

        response = hook.publish_to_target(target, message)

        assert "MessageId" in response

    def test_publish_to_target_error(self):
        hook = SnsHook(aws_conn_id="aws_default")

        message = "Hello world"
        topic_name = "test-topic"
        target = hook.get_conn().create_topic(Name=topic_name).get("TopicArn")

        error_message = (
            r"Values in MessageAttributes must be one of bytes, str, int, float, or iterable; got .*"
        )
        with pytest.raises(TypeError, match=error_message):
            hook.publish_to_target(
                target,
                message,
                message_attributes={
                    "test-non-iterable": object(),
                },
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
AWS ECS Executor Utilities.

Data classes and utility functions used by the ECS executor.
"""

from __future__ import annotations

import datetime
from collections import defaultdict
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable

from inflection import camelize

from airflow.providers.amazon.aws.executors.utils.base_config_keys import BaseConfigKeys
from airflow.utils.state import State

if TYPE_CHECKING:
    from airflow.models.taskinstance import TaskInstanceKey

CommandType = Sequence[str]
ExecutorConfigFunctionType = Callable[[CommandType], dict]
ExecutorConfigType = dict[str, Any]

ECS_LAUNCH_TYPE_EC2 = "EC2"
ECS_LAUNCH_TYPE_FARGATE = "FARGATE"

CONFIG_GROUP_NAME = "aws_ecs_executor"

CONFIG_DEFAULTS = {
    "conn_id": "aws_default",
    "max_run_task_attempts": "3",
    "assign_public_ip": "False",
    "platform_version": "LATEST",
    "check_health_on_startup": "True",
}


@dataclass
class EcsQueuedTask:
    """Represents an ECS task that is queued. The task will be run in the next heartbeat."""

    key: TaskInstanceKey
    command: CommandType
    queue: str
    executor_config: ExecutorConfigType
    attempt_number: int
    next_attempt_time: datetime.datetime


@dataclass
class EcsTaskInfo:
    """Contains information about a currently running ECS task."""

    cmd: CommandType
    queue: str
    config: ExecutorConfigType


class RunTaskKwargsConfigKeys(BaseConfigKeys):
    """Keys loaded into the config which are valid ECS run_task kwargs."""

    ASSIGN_PUBLIC_IP = "assign_public_ip"
    CAPACITY_PROVIDER_STRATEGY = "capacity_provider_strategy"
    CLUSTER = "cluster"
    CONTAINER_NAME = "container_name"
    LAUNCH_TYPE = "launch_type"
    PLATFORM_VERSION = "platform_version"
    SECURITY_GROUPS = "security_groups"
    SUBNETS = "subnets"
    TASK_DEFINITION = "task_definition"


class AllEcsConfigKeys(RunTaskKwargsConfigKeys):
    """All keys loaded into the config which are related to the ECS Executor."""

    AWS_CONN_ID = "conn_id"
    CHECK_HEALTH_ON_STARTUP = "check_health_on_startup"
    MAX_RUN_TASK_ATTEMPTS = "max_run_task_attempts"
    REGION_NAME = "region_name"
    RUN_TASK_KWARGS = "run_task_kwargs"


class EcsExecutorException(Exception):
    """Thrown when something unexpected has occurred within the ECS ecosystem."""


class EcsExecutorTask:
    """Data Transfer Object for an ECS Task."""

    def __init__(
        self,
        task_arn: str,
        last_status: str,
        desired_status: str,
        containers: list[dict[str, Any]],
        started_at: Any | None = None,
        stopped_reason: str | None = None,
        external_executor_id: str | None = None,
    ):
        self.task_arn = task_arn
        self.last_status = last_status
        self.desired_status = desired_status
        self.containers = containers
        self.started_at = started_at
        self.stopped_reason = stopped_reason
        self.external_executor_id = external_executor_id

    def get_task_state(self) -> str:
        """
        Determine the state of an ECS task based on its status and other relevant attributes.

        It can return one of the following statuses:
            QUEUED - Task is being provisioned.
            RUNNING - Task is launched on ECS.
            REMOVED - Task provisioning has failed for some reason. See `stopped_reason`.
            FAILED - Task is completed and at least one container has failed.
            SUCCESS - Task is completed and all containers have succeeded.
        """
        if self.last_status == "RUNNING":
            return State.RUNNING
        if self.desired_status == "RUNNING":
            return State.QUEUED
        is_finished = self.desired_status == "STOPPED"
        has_exit_codes = all(["exit_code" in x for x in self.containers])
        # Sometimes ECS tasks may time out.
        if not self.started_at and is_finished:
            return State.REMOVED
        if not is_finished or not has_exit_codes:
            return State.RUNNING
        all_containers_succeeded = all([x["exit_code"] == 0 for x in self.containers])
        return State.SUCCESS if all_containers_succeeded else State.FAILED

    def __repr__(self):
        """Return a string representation of the ECS task."""
        return f"({self.task_arn}, {self.last_status}->{self.desired_status}, {self.get_task_state()})"


class EcsTaskCollection:
    """A five-way dictionary between Airflow task ids, Airflow cmds, ECS ARNs, and ECS task objects."""

    def __init__(self):
        self.key_to_arn: dict[TaskInstanceKey, str] = {}
        self.arn_to_key: dict[str, TaskInstanceKey] = {}
        self.tasks: dict[str, EcsExecutorTask] = {}
        self.key_to_failure_counts: dict[TaskInstanceKey, int] = defaultdict(int)
        self.key_to_task_info: dict[TaskInstanceKey, EcsTaskInfo] = {}

    def add_task(
        self,
        task: EcsExecutorTask,
        airflow_task_key: TaskInstanceKey,
        queue: str,
        airflow_cmd: CommandType,
        exec_config: ExecutorConfigType,
        attempt_number: int,
    ):
        """Add a task to the collection."""
        arn = task.task_arn
        self.tasks[arn] = task
        self.key_to_arn[airflow_task_key] = arn
        self.arn_to_key[arn] = airflow_task_key
        self.key_to_task_info[airflow_task_key] = EcsTaskInfo(airflow_cmd, queue, exec_config)
        self.key_to_failure_counts[airflow_task_key] = attempt_number

    def update_task(self, task: EcsExecutorTask):
        """Update the state of the given task based on task ARN."""
        self.tasks[task.task_arn] = task

    def task_by_key(self, task_key: TaskInstanceKey) -> EcsExecutorTask:
        """Get a task by Airflow Instance Key."""
        arn = self.key_to_arn[task_key]
        return self.task_by_arn(arn)

    def task_by_arn(self, arn) -> EcsExecutorTask:
        """Get a task by AWS ARN."""
        return self.tasks[arn]

    def pop_by_key(self, task_key: TaskInstanceKey) -> EcsExecutorTask:
        """Delete task from collection based off of Airflow Task Instance Key."""
        arn = self.key_to_arn[task_key]
        task = self.tasks[arn]
        del self.key_to_arn[task_key]
        del self.key_to_task_info[task_key]
        del self.arn_to_key[arn]
        del self.tasks[arn]
        if task_key in self.key_to_failure_counts:
            del self.key_to_failure_counts[task_key]
        return task

    def get_all_arns(self) -> list[str]:
        """Get all AWS ARNs in collection."""
        return list(self.key_to_arn.values())

    def get_all_task_keys(self) -> list[TaskInstanceKey]:
        """Get all Airflow Task Keys in collection."""
        return list(self.key_to_arn.keys())

    def failure_count_by_key(self, task_key: TaskInstanceKey) -> int:
        """Get the number of times a task has failed given an Airflow Task Key."""
        return self.key_to_failure_counts[task_key]

    def increment_failure_count(self, task_key: TaskInstanceKey):
        """Increment the failure counter given an Airflow Task Key."""
        self.key_to_failure_counts[task_key] += 1

    def info_by_key(self, task_key: TaskInstanceKey) -> EcsTaskInfo:
        """Get the Airflow Command given an Airflow task key."""
        return self.key_to_task_info[task_key]

    def __getitem__(self, value):
        """Get a task by AWS ARN."""
        return self.task_by_arn(value)

    def __len__(self):
        """Determine the number of tasks in collection."""
        return len(self.tasks)


def _recursive_flatten_dict(nested_dict):
    """
    Recursively unpack a nested dict and return it as a flat dict.

    For example, _flatten_dict({'a': 'a', 'b': 'b', 'c': {'d': 'd'}}) returns {'a': 'a', 'b': 'b', 'd': 'd'}.
    """
    items = []
    for key, value in nested_dict.items():
        if isinstance(value, dict):
            items.extend(_recursive_flatten_dict(value).items())
        else:
            items.append((key, value))
    return dict(items)


def parse_assign_public_ip(assign_public_ip, is_launch_type_ec2=False):
    """Convert "assign_public_ip" from True/False to ENABLE/DISABLE."""
    # If the launch type is EC2, you cannot/should not provide the assignPublicIp parameter (which is
    # specific to Fargate)
    if not is_launch_type_ec2:
        return "ENABLED" if assign_public_ip == "True" else "DISABLED"


def camelize_dict_keys(nested_dict) -> dict:
    """Accept a potentially nested dictionary and recursively convert all keys into camelCase."""
    result = {}
    for key, value in nested_dict.items():
        new_key = camelize(key, uppercase_first_letter=False)
        if isinstance(value, dict) and (key.lower() != "tags"):
            # The key name on tags can be whatever the user wants, and we should not mess with them.
            result[new_key] = camelize_dict_keys(value)
        else:
            result[new_key] = nested_dict[key]
    return result



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import timedelta
from pathlib import Path
from unittest import mock
from unittest.mock import MagicMock, patch

import pytest
from botocore.waiter import Waiter
from jinja2 import StrictUndefined

from airflow.exceptions import TaskDeferred
from airflow.models import DAG, DagRun, TaskInstance
from airflow.providers.amazon.aws.operators.emr import EmrCreateJobFlowOperator
from airflow.providers.amazon.aws.triggers.emr import EmrCreateJobFlowTrigger
from airflow.providers.amazon.aws.utils.waiter import WAITER_POLICY_NAME_MAPPING, WaitPolicy
from airflow.utils import timezone
from airflow.utils.state import DagRunState
from airflow.utils.types import DagRunType

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS
from unit.amazon.aws.utils.test_template_fields import validate_template_fields
from unit.amazon.aws.utils.test_waiter import assert_expected_waiter_type

TASK_ID = "test_task"

TEST_DAG_ID = "test_dag_id"

DEFAULT_DATE = timezone.datetime(2017, 1, 1)

JOB_FLOW_ID = "j-8989898989"
RUN_JOB_FLOW_SUCCESS_RETURN = {"ResponseMetadata": {"HTTPStatusCode": 200}, "JobFlowId": JOB_FLOW_ID}

TEMPLATE_SEARCHPATH = Path(__file__).parents[1].joinpath("config_templates").as_posix()


@pytest.fixture
def mocked_hook_client():
    with patch("airflow.providers.amazon.aws.hooks.emr.EmrHook.conn") as m:
        yield m


class TestEmrCreateJobFlowOperator:
    # When
    _config = {
        "Name": "test_job_flow",
        "ReleaseLabel": "5.11.0",
        "Steps": [
            {
                "Name": "test_step",
                "ActionOnFailure": "CONTINUE",
                "HadoopJarStep": {
                    "Jar": "command-runner.jar",
                    "Args": ["/usr/lib/spark/bin/run-example", "{{ macros.ds_add(ds, -1) }}", "{{ ds }}"],
                },
            }
        ],
    }

    def setup_method(self):
        args = {"owner": "airflow", "start_date": DEFAULT_DATE}
        self.operator = EmrCreateJobFlowOperator(
            task_id=TASK_ID,
            aws_conn_id="aws_default",
            emr_conn_id="emr_default",
            region_name="ap-southeast-2",
            dag=DAG(
                TEST_DAG_ID,
                schedule=None,
                default_args=args,
                template_searchpath=TEMPLATE_SEARCHPATH,
                template_undefined=StrictUndefined,
            ),
        )
        self.mock_context = MagicMock()

    def test_init(self):
        assert self.operator.aws_conn_id == "aws_default"
        assert self.operator.emr_conn_id == "emr_default"
        assert self.operator.region_name == "ap-southeast-2"

    @pytest.mark.db_test
    def test_render_template(self, session, clean_dags_and_dagruns):
        self.operator.job_flow_overrides = self._config
        if AIRFLOW_V_3_0_PLUS:
            dag_run = DagRun(
                dag_id=self.operator.dag_id,
                logical_date=DEFAULT_DATE,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        else:
            dag_run = DagRun(
                dag_id=self.operator.dag_id,
                execution_date=DEFAULT_DATE,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        ti = TaskInstance(task=self.operator)
        ti.dag_run = dag_run
        session.add(ti)
        session.commit()
        ti.render_templates()

        expected_args = {
            "Name": "test_job_flow",
            "ReleaseLabel": "5.11.0",
            "Steps": [
                {
                    "Name": "test_step",
                    "ActionOnFailure": "CONTINUE",
                    "HadoopJarStep": {
                        "Jar": "command-runner.jar",
                        "Args": [
                            "/usr/lib/spark/bin/run-example",
                            (DEFAULT_DATE - timedelta(days=1)).strftime("%Y-%m-%d"),
                            DEFAULT_DATE.strftime("%Y-%m-%d"),
                        ],
                    },
                }
            ],
        }

        assert self.operator.job_flow_overrides == expected_args

    @pytest.mark.db_test
    def test_render_template_from_file(self, mocked_hook_client, session, clean_dags_and_dagruns):
        self.operator.job_flow_overrides = "job.j2.json"
        self.operator.params = {"releaseLabel": "5.11.0"}

        if AIRFLOW_V_3_0_PLUS:
            dag_run = DagRun(
                dag_id=self.operator.dag_id,
                logical_date=DEFAULT_DATE,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        else:
            dag_run = DagRun(
                dag_id=self.operator.dag_id,
                execution_date=DEFAULT_DATE,
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        ti = TaskInstance(task=self.operator)
        ti.dag_run = dag_run
        session.add(ti)
        session.commit()
        ti.render_templates()

        mocked_hook_client.run_job_flow.return_value = RUN_JOB_FLOW_SUCCESS_RETURN

        # String in job_flow_overrides (i.e. from loaded as a file) is not "parsed" until inside execute()
        self.operator.execute(self.mock_context)

        expected_args = {
            "Name": "test_job_flow",
            "ReleaseLabel": "5.11.0",
            "Steps": [
                {
                    "Name": "test_step",
                    "ActionOnFailure": "CONTINUE",
                    "HadoopJarStep": {
                        "Jar": "command-runner.jar",
                        "Args": [
                            "/usr/lib/spark/bin/run-example",
                            "2016-12-31",
                            "2017-01-01",
                        ],
                    },
                }
            ],
        }

        assert self.operator.job_flow_overrides == expected_args

    def test_execute_returns_job_id(self, mocked_hook_client):
        mocked_hook_client.run_job_flow.return_value = RUN_JOB_FLOW_SUCCESS_RETURN
        assert self.operator.execute(self.mock_context) == JOB_FLOW_ID

    @pytest.mark.parametrize(
        "wait_policy",
        [
            pytest.param(WaitPolicy.WAIT_FOR_COMPLETION, id="with wait for completion"),
            pytest.param(WaitPolicy.WAIT_FOR_STEPS_COMPLETION, id="with wait for steps completion policy"),
        ],
    )
    @mock.patch("botocore.waiter.get_service_module_name", return_value="emr")
    @mock.patch.object(Waiter, "wait")
    def test_execute_with_wait_policy(self, mock_waiter, _, mocked_hook_client, wait_policy: WaitPolicy):
        mocked_hook_client.run_job_flow.return_value = RUN_JOB_FLOW_SUCCESS_RETURN

        # Mock out the emr_client creator
        self.operator.wait_policy = wait_policy

        assert self.operator.execute(self.mock_context) == JOB_FLOW_ID
        mock_waiter.assert_called_once_with(mock.ANY, ClusterId=JOB_FLOW_ID, WaiterConfig=mock.ANY)
        assert_expected_waiter_type(mock_waiter, WAITER_POLICY_NAME_MAPPING[wait_policy])

    def test_create_job_flow_deferrable(self, mocked_hook_client):
        """
        Test to make sure that the operator raises a TaskDeferred exception
        if run in deferrable mode.
        """
        mocked_hook_client.run_job_flow.return_value = RUN_JOB_FLOW_SUCCESS_RETURN

        self.operator.deferrable = True
        with pytest.raises(TaskDeferred) as exc:
            self.operator.execute(self.mock_context)

        assert isinstance(exc.value.trigger, EmrCreateJobFlowTrigger), (
            "Trigger is not a EmrCreateJobFlowTrigger"
        )

    def test_template_fields(self):
        validate_template_fields(self.operator)
